/*
 * Copyright (C) 2002-2012 OpenClovis Solutions Inc.  All Rights Reserved.
 *
 * This file is available  under  a  commercial  license  from  the
 * copyright  holder or the GNU General Public License Version 2.0.
 * 
 * The source code for  this program is not published  or otherwise 
 * divested of  its trade secrets, irrespective  of  what  has been 
 * deposited with the U.S. Copyright office.
 * 
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 * 
 * For more  information, see  the file  COPYING provided with this
 * material.
 */
/*******************************************************************************
 * ModuleName  : cm                                                            
 * File        : clCmApis.c
 *******************************************************************************/

/*******************************************************************************
 * Description :                                                                
 * This file contains the implementation of client side apis of Chassis   
 * Manager Module.	
 *******************************************************************************/

#undef __SERVER__
#define __CLIENT__

#include <clChassisMgrApi.h>
#include <clCmDefs.h>
#include <clIocApi.h>
#include <clIocIpi.h>
#include <clIocServices.h>
#include <clCorApi.h>
#include <clCorUtilityApi.h>
#include <clDebugApi.h>
#include <clCommonErrors.h>
#include <string.h>
#include <clChassisMgrApi.h>
#include <clCpmExtApi.h>
#include <clCmServerFuncTable.h>
#include <clXdrApi.h>
#include "clCmApis.h"

/* The fallowing macro is defined to deferentiate the existing CM and multi chassis CM.
   This macro value should not conflict with "flag" values in rmdApi.h
*/
#define CL_CM__CHASSIS_PARAM_DELIMITER ","

ClUint16T  clCmMaxChassis = 0;
ClUint16T  clCmMaxPhysicalSlots = 0;
ClUint16T  clCmMaxAMCSlots = 0;

ClUint16T clCmMaxChassisBits = 0;
ClUint16T clCmMaxPhysicalBits = 0;
ClUint16T clCmMaxAMCBits = 0;

static void clCmEnvGet();

static ClRcT clCmServerAddressGet(ClIocPhysicalAddressT *pIocAddress);

static ClRcT clCmRmdWithMsg(ClUint32T funcId, ClBufferHandleT inMsgHdl, ClBufferHandleT outMsgHdl, ClUint32T flags, ClRmdAsyncOptionsT *pAsyncOptions);

static ClRcT clCmSlotIdFromMOIdGet(CL_IN ClCorMOIdPtrT hMoId, CL_OUT ClUint32T * pChassisId, CL_OUT ClUint32T * pSlotId);

ClRcT clCmLibInitialize(void)
{
    ClRcT rc = CL_OK;
    
    clCmEnvGet();
    
    rc = clEoClientTableRegister(CL_EO_CLIENT_SYM_MOD(gAspFuncTable, CM), CL_IOC_CM_PORT);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_ERROR,"CM","INI","Initialization of CM Library failed, Err Code:[%d]",rc);
        return rc;
    }

    clLog(CL_LOG_SEV_DEBUG,"CM","INI","Successfully Initialized CM Library");
    return rc;
}

ClRcT clCmLibFinalize(void)
{
    /* empty function to satisfy linker. */
    return CL_OK;
}

static ClUint16T clCmGetNumofBits(ClUint16T size)
{
    ClUint16T  order = 0;
    do
    {
        ++order;
        size >>= 1;
    } while(size >  0);
    return order;
}

static void clCmEnvGet()
{
    ClCharT  *pEnvVar = NULL;
    ClCharT  pEnvVarLocal[CL_MAX_NAME_LENGTH] = {0};
    /* Handling of muiltiple chassis */
    if(NULL == (pEnvVar = getenv("CL_CM_MULTI_CHASSIS_PARAM")))
    {
        clLog(CL_LOG_SEV_NOTICE, "ENV", "VAR", " CM handles Only one CHASSIS (default)");
    }
    else
    {
        ClCharT *token;
        strcpy(pEnvVarLocal,pEnvVar);
        clLog(CL_LOG_SEV_DEBUG, "ENV", "VAR", " CL_CM_MULTI_CHASSIS_PARAM = %s", pEnvVarLocal);
        token = strtok(pEnvVarLocal, CL_CM__CHASSIS_PARAM_DELIMITER);
        if(token != NULL ) 
        {
            clCmMaxChassis = atoi(token);
            clLog(CL_LOG_SEV_INFO, "ENV", "VAR", "Max Chassis [%u]", clCmMaxChassis);
            token = strtok(NULL, CL_CM__CHASSIS_PARAM_DELIMITER);
            if(token != NULL)
            {
                clCmMaxPhysicalSlots = atoi(token);
                clLog(CL_LOG_SEV_INFO, "ENV", "VAR", "Max Physical Slots [%u]", clCmMaxPhysicalSlots);
                token = strtok(NULL, CL_CM__CHASSIS_PARAM_DELIMITER);
                if( token != NULL)
                {
                    clCmMaxAMCSlots = atoi(token);
                    clLog(CL_LOG_SEV_INFO, "ENV", "VAR", "Max AMC slots [%u]", clCmMaxAMCSlots);
                } 
            }
        }
    }
    if( clCmMaxChassis > 1)
        clCmMaxChassisBits = clCmGetNumofBits(clCmMaxChassis-1);
    else if (clCmMaxChassis == 1)
        clCmMaxChassisBits = 1; 
    clCmMaxPhysicalBits = CL_CM_MAX_NODE_BITS - clCmMaxChassisBits;
    if (clCmMaxPhysicalSlots && clCmMaxAMCSlots) {
        clCmMaxPhysicalBits = clCmGetNumofBits(clCmMaxPhysicalSlots);
    }
    clCmMaxAMCBits = 0;
    if (clCmMaxAMCSlots)
    {
        ClUint16T maxbits = clCmGetNumofBits(clCmMaxAMCSlots);
        clCmMaxAMCBits = CL_CM_MAX_NODE_BITS - clCmMaxChassisBits - clCmMaxPhysicalBits;
        if (maxbits > clCmMaxAMCBits)
        {
            /* Error Configuration in in Multi chassis parameters */
           clLog(CL_LOG_SEV_ERROR, "ENV", "VAR", "Error in Multi chassis Configuation");
           exit(0);
        }
    }
  
}    
static ClRcT clCmSlotIdFromMOIdGet(CL_IN ClCorMOIdPtrT hMoId, CL_OUT ClUint32T *pChassisId, CL_OUT ClUint32T *pSlotId)
{
    ClCpmSlotInfoT cpmSlotInfo; 
    ClRcT rc=CL_OK;
    ClUint32T IocAddress;

    if(pChassisId==NULL || pSlotId==NULL)
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED," NULL POINTER passed in clCmSlotIdFromMOIdGet");
        return CL_ERR_NULL_POINTER;
    }

    memset(&cpmSlotInfo,0,sizeof(ClCpmSlotInfoT )); 
    memcpy(&(cpmSlotInfo.nodeMoId), hMoId,sizeof(ClCorMOIdT));

    rc=clCpmSlotInfoGet(CL_CPM_NODE_MOID, & cpmSlotInfo);
    if(CL_OK!=rc)
    {
        *pSlotId=0;
        *pChassisId=0;
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED," clCpmSlotInfoGet Failed rc:%x",rc);
    }
    else 
    {
        
        IocAddress = cpmSlotInfo.nodeIocAddress;
        
        *pSlotId = ( IocAddress >> clCmMaxAMCBits) & PHYSICAL_SLOT_MASk  ;
        
        *pChassisId = (IocAddress >> (clCmMaxAMCBits+clCmMaxPhysicalBits)) & CHASSIS_ID_MASK; 
    }
    return rc;
}

static ClRcT clCmRmdWithMsg(ClUint32T funcId, ClBufferHandleT inMsgHdl, ClBufferHandleT outMsgHdl, ClUint32T flags, ClRmdAsyncOptionsT *pAsyncOptions)
{
    ClRcT rc = CL_OK;
    ClRmdOptionsT rmdOptions = CL_RMD_DEFAULT_OPTIONS;
    ClUint8T status = 0;
    ClIocAddressT cmServerAddress;

    rc=clCmServerAddressGet(&cmServerAddress.iocPhyAddress);
    if(rc != CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, " clCmServerAddressGet Failed rc:0x%x",rc);
        return rc;
    }
    CL_CM_CLIENT_DEBUG_PRINT(("CmServer IocNode Address =%d port Id = %d", cmServerAddress.iocPhyAddress.nodeAddress, cmServerAddress.iocPhyAddress.portId));
    
    clIocCompStatusGet(cmServerAddress.iocPhyAddress, &status);
    if(status == CL_IOC_NODE_DOWN)
    {
        return CL_IOC_RC(CL_IOC_ERR_COMP_UNREACHABLE);
    }

    memset((void*)&rmdOptions, 0 ,sizeof(rmdOptions));

    rmdOptions.timeout = CL_RMD_DEFAULT_TIMEOUT ;
    rmdOptions.priority = CL_RMD_DEFAULT_PRIORITY ;
    rmdOptions.retries = CL_RMD_DEFAULT_RETRIES;
    rc = clRmdWithMsg(cmServerAddress, funcId, inMsgHdl, (ClBufferHandleT)outMsgHdl, flags, &rmdOptions, pAsyncOptions);
    if(rc != CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, " RMD Call in to CM failed with rc:0x%x",rc);
    }
    return rc;

}

/**Description:
 * ------------
 * This API is used to get the state of the FRU . The states are defined in
 * the SaHpi.h header. First argument is the moid of the FRU for which the
 * state to be get. Second argument is a pointer to location to store the
 * state of the FRU. 
 *
 * @param in hMoId	:	moid of the FRU for which state to be Get.
 * @param out pState:	Pointer to store the state of the FRU. 
 *
 * @returns CL_OK           :  on success full retrieval of the state.
 *			CL_ERR_INVALID_PARAMETER :  If any input parameter is invalid.
 * 			CL_ERR_NULL_POINTER    :  If NULL is passed to store the state.
 *			CL_ERR_VERSION_MISMATCH : If client and server are version
 *									  incompatible.
 */

ClRcT clCmFruStateGet (ClCorMOIdPtrT hMoId,SaHpiHsStateT *pState)
{
    ClRcT rc = CL_OK;
    ClUint32T flags = CL_RMD_CALL_ATMOST_ONCE | CL_RMD_CALL_NEED_REPLY; 
    ClBufferHandleT inputMsg , outputMsg;
    ClUint32T outputLen = (ClUint32T)sizeof(SaHpiHsStateT);
    ClCmFruAddressT fruSlotInfo;
    CL_CM_VERSION_SET(fruSlotInfo.version);

    if((rc = clCorMoIdValidate(hMoId))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Invalid MOID passed ");
	return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }

    if(pState == NULL )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, " NULL passed for storing state");
	return CL_CM_ERROR(CL_ERR_NULL_POINTER);
    }


    /* FIXME : As of now we assume the physical slot is same as the logical
     *  slot , Need to fix this later */
    rc = clCmSlotIdFromMOIdGet(hMoId, &(fruSlotInfo.chassisId), &(fruSlotInfo.physicalSlot));
    if(rc != CL_OK)
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Getting Slot id from moid failed rc:0x%x",rc);
	return rc;
    }

    /* form the message with api request */ 
    if((rc = clBufferCreate( &inputMsg ))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Message Creation failed rc:0x%x",rc);
	return rc;
    }

    if((rc = clBufferCreate( &outputMsg ))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Message Creation failed rc:0x%x",rc);
	clBufferDelete(&inputMsg );
	return rc;
    }

    /* copy the api req info to the Message */ 
    /* Mahesh. changed the message to include Moid
     *  instead of structure having unions */
    clBufferNBytesWrite(inputMsg , (ClUint8T *)&fruSlotInfo, (ClUint32T)sizeof(ClUint32T));
    
    if((rc = clCmRmdWithMsg(CL_CHASSIS_MANAGER_FRU_STATE_GET, inputMsg, outputMsg, flags, NULL))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "clCmRmdWithMsg Call in to CM failed with rc:0x%x",rc);
	clBufferDelete(&inputMsg);
	clBufferDelete(&outputMsg);
	return rc;
    }

    clBufferNBytesRead(outputMsg, (ClUint8T*)pState, (ClUint32T*)&outputLen);

    clBufferDelete(&inputMsg);
    clBufferDelete(&outputMsg);

    return rc;
}


/** Description:
 *	------------
 * This api is used to operate on the FRU or any platform hardware.API takes 2 
 * arguments , first argument is the moid of the FRU to operate , second
 * argument is the operation requested. Operations are listed in the
 * enumeration cmRequestT. 
 *
 * @param hMoId				:	moid of the platform hardware to operate.
 * @param request 			:       Operation requested on the hardware.
 * 
 * @returns	CL_OK 		   : 	on successfull return.
 * 			CL_ERR_INVALID_PARAMETER :	If any of the input arguments is invalid.
 *			CL_ERR_VERSION_MISMATCH : If client and server are version
 *									  incompatible.
 *
 */
ClRcT clCmFruOperationRequest (ClCorMOIdPtrT hMoId, ClCmFruOperationT request)
{

    ClRcT rc = CL_OK;
    ClCmOpReqT cmOpReq;

    if(( rc = clCorMoIdValidate(hMoId))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Invalid MOID passed ");
	return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }
    if(request < CL_CM_POWERON_REQUEST || request > CL_CM_RESET_DEASSERT_REQUEST )
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Invalid Request passed ");
	return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }

    memset((void*)&cmOpReq, 0 , sizeof(ClCmOpReqT));
    /* FIXME : As of now we assume the physical slot is same as the logical
     *  slot , Need to fix this later */
    rc = clCmSlotIdFromMOIdGet(hMoId, &(cmOpReq.chassisId), &(cmOpReq.physicalSlot));
    if(rc != CL_OK)
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Getting Slot id from moid failed rc:0x%x",rc);
	return rc;
    }

    rc=clCmBladeOperationRequest (cmOpReq.chassisId, cmOpReq.physicalSlot, request);
    /* form the message with api request */ 
    if(rc != CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"clCmBladeOperationRequest failed rc:0x%x",rc);
    }
    return rc;
}

ClRcT clCmBladeOperationRequest (ClUint32T chassisId, ClUint32T slotId, ClCmFruOperationT request)
{
    ClRcT rc = CL_OK;
    ClUint32T flags = CL_RMD_CALL_ATMOST_ONCE;
    ClCmOpReqT bladeOpReq;
    ClBufferHandleT inputMsg;

    if(request < CL_CM_POWERON_REQUEST || request > CL_CM_RESET_DEASSERT_REQUEST )
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Invalid Request passed ");
        return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }
    memset((void*)&bladeOpReq, 0 , sizeof(bladeOpReq));
    CL_CM_VERSION_SET(bladeOpReq.version); 

    bladeOpReq.chassisId = chassisId;
    bladeOpReq.physicalSlot = slotId;
    bladeOpReq.request =  request;

    CL_CM_CLIENT_DEBUG_PRINT(("chassisId=%d,slotId=%d,request=%d",chassisId,slotId,request));

    /* form the message with api request */ 
    if((rc = clBufferCreate( &inputMsg ))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Message Creation failed rc:0x%x",rc);
	return rc;
    }

    /* copy the blade req info to the Message */ 
    rc=clBufferNBytesWrite( inputMsg , (ClUint8T *)&bladeOpReq, (ClUint32T)sizeof(bladeOpReq));
    if(rc != CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"clBufferNBytesWrite Failed rc:0x%x",rc);
	clBufferDelete(&inputMsg);
	return rc;
    }
   
    if((rc = clCmRmdWithMsg(CL_CHASSIS_MANAGER_OPERATION_REQUEST, inputMsg, (ClBufferHandleT)NULL, flags, NULL))!= CL_OK )
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"clCmRmdWithMsg Call in to CM failed with rc:0x%x",rc);
	clBufferDelete(&inputMsg );
	return rc;
    }

    clBufferDelete(&inputMsg);
    return rc;
}



static ClRcT clCmServerAddressGet(ClIocPhysicalAddressT *pIocAddress)
{
    ClRcT rc=CL_OK;

    /*FIXME: Need to use the name service apis to resolve the CM address for
     * now assuming that the cm will be colocated with Masters*/
    rc=clCpmMasterAddressGet(&pIocAddress->nodeAddress); 
    if(rc != CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"clCpmMasterAddressGet failed rc:0x%x",rc);
    }
    else
    {
        clLog(CL_LOG_SEV_INFO, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"cm nodeAddress = %x",pIocAddress->nodeAddress);
        pIocAddress->portId = CL_IOC_CM_PORT;
    }
    return rc ;
}

/**
 *  Determine the supported cm Version.  
 *
 *  This Api checks the version is supported or not. If the version check fails,
 *	the closest version is passed as the out parameter.
 *                                                                        
 *  @param version[IN/OUT] : version value.
 *
 *  @returns CL_OK  - Success<br>
 *    On failure, the error CL_ERR_VERSION_MISMATCH is returned with the 
 * 	  closest value of the version supported filled in the version as out parameter.
 */
ClRcT clCmVersionVerify(ClVersionT *version)
{
    ClRcT rc = CL_OK;
    if(version == NULL)
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"NULL pointer passed as version address.with rc: 0x%x", rc);
	return CL_ERR_NULL_POINTER;
    }
    CL_FUNC_ENTER();
    
    if((rc = clVersionVerify(&gCmClientToServerVersionDb, version)))
    {
	clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"The CM version is not supported. 0x%x", rc);
	return CL_ERR_VERSION_MISMATCH;
    }
    CL_FUNC_EXIT();
    return rc;
}

ClRcT clCmThresholdStateGet(ClUint32T slot, ClCmThresholdLevelT *pLevel, ClBoolT *pAsserted)
{
    ClRcT rc = CL_OK;
    ClUint32T flags = CL_RMD_CALL_ATMOST_ONCE | CL_RMD_CALL_NEED_REPLY;
    ClCmThresholdStateReqT stateReq = {{0}};
    SaHpiEventStateT eventState = 0;
    ClBufferHandleT inMsg = 0;
    ClBufferHandleT outMsg = 0;

    if(!slot || (!pLevel && !pAsserted))
    {
        return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }

    CL_CM_VERSION_SET(stateReq.version);
    stateReq.slot = slot;
    stateReq.sensorNum = 0;

    rc = clBufferCreate(&inMsg);
    rc |= clBufferCreate(&outMsg);

    if(rc != CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Message Creation failed rc:0x%x",rc);
        goto out_delete;
    }
    
    rc =  clXdrMarshallClVersionT((void*)&stateReq.version, inMsg, 0);
    rc |= clXdrMarshallClUint32T((void*)&stateReq.slot, inMsg, 0);
    rc |= clXdrMarshallClUint32T((void*)&stateReq.sensorNum, inMsg, 0);
    
    if(rc != CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Xdr marshall of threshold request failed with [%#x]", rc);
        goto out_delete;
    }
    
    if((rc = clCmRmdWithMsg(CL_CHASSIS_MANAGER_THRESHOLD_STATE_GET, inMsg, outMsg, flags, NULL))!= CL_OK )
    {
       if(CL_GET_ERROR_CODE(rc) != CL_IOC_ERR_COMP_UNREACHABLE)
       {
           clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"clCmRmdWithMsg Call in to CM failed with rc:0x%x",rc);
       }
       goto out_delete;
    }

    rc = clXdrUnmarshallClUint16T(outMsg, (void*)&eventState);
    if(rc != CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Umarshall of event state failed with [%#x]", rc);
        goto out_delete;
    }

    if(pLevel)
        *pLevel = (ClCmThresholdLevelT)eventState;
    
    if(pAsserted)
    {
        *pAsserted = CL_FALSE;
        if((eventState & (CL_CM_THRESHOLD_LOWER_MAJOR | CL_CM_THRESHOLD_LOWER_CRIT | \
                          CL_CM_THRESHOLD_UPPER_MAJOR | CL_CM_THRESHOLD_UPPER_CRIT)))
           *pAsserted = CL_TRUE;
    }

    out_delete:
    if(inMsg) clBufferDelete(&inMsg);
    if(outMsg) clBufferDelete(&outMsg);
    return rc;
}
