/* 
 * The source code for  this program is not published  or otherwise 
 * divested of  its trade secrets, irrespective  of  what  has been 
 * deposited with the U.S. Copyright office.
 * 
 * This program is  free software; you can redistribute it and / or
 * modify  it under  the  terms  of  the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 * 
 * You  should  have  received  a  copy of  the  GNU General Public
 * License along  with  this program. If  not,  write  to  the 
 * Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * Build: 4.2.0
 */
/*******************************************************************************
 * ModuleName  : cm
 * File    : clCmHpiInterface.c
 *******************************************************************************/

/*******************************************************************************
 * Description :
 * This file contains the interface implementation of the chassis manager 
 * with HPI.                                  
 **************************************************************************/
/*Wraper header addition */
#include <string.h>
#include <unistd.h>

#include <SaHpi.h>
#include <oh_utils.h>       /* From OpenHPI package */


#include <clCommon.h>
#include <clDebugApi.h>
#include <clAlarmDefinitions.h>
#include <clOsalApi.h>
#include <clCorApi.h>
#include <clCorUtilityApi.h>
#include <clLogApi.h>
#include <clCpmExtApi.h>
#include <clCmDefs.h>
#include <clHash.h>
#include <clList.h>

#include "clCmHpiDebug.h"
#include "clCmUtils.h"
#include <clCmPluginApi.h>
#include <clHpiEvtInterface.h>

#define AREA_HPI    "HPI"
#define CTX_INI     "INI"   /* Initialization context */
#define CTX_EVT     "EVT"   /* HPI event handler context */
#define CTX_DIS     "DIS"   /* discovery context */
#define CTX_REQ     "REQ"   /* client request context */
#define _OPENHPI_SESSION_RETRIES_DEFAULT (100000)


/* Comment the following line after testing HPI CHANNEL event simulation */
#undef HPI_CHANNEL_EVENT_SIM

static ClRcT clCmDomainTaskCreate(SaHpiDomainIdT domain_id);
static ClRcT clCmDomainTaskDelete(SaHpiDomainIdT domain_id);
static ClRcT DiscoverPlatform ( SaHpiSessionIdT  session_id, SaHpiDomainIdT domain_id, ClUint8T *domainResourceDiscovery );
ClCmDomainListT* clCmDomainEntryGet( SaHpiDomainIdT domainId);
static ClRcT hpiEventProcessingLoop(void *ptr);
static ClRcT clCmDomainThread(void *ptr);
static ClRcT openHpiSession ( SaHpiSessionIdT *session_id, SaHpiDomainIdT domain_id );
static ClRcT clCmCreateAndPublishResHsEvt(SaHpiSessionIdT sessionid, SaHpiRptEntryT *pRptEntry, SaHpiHsStateT  prevHsState, SaHpiHsStateT  curHsState);
static ClRcT clCmCreateResourceHotSwapEvent(SaHpiSessionIdT sessionid, SaHpiRptEntryT *pRptEntry, SaHpiHsStateT  prevHsState,
                                            SaHpiHsStateT  curHsState, ClEventT *HotSwapEvt);
static ClRcT resourceHotSwapEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT * );
static ClRcT resourceHotSwapEventInformAmf(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT * );
static ClRcT clCmSensorEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT  *rptentry, SaHpiRdrT *pRdr);
static ClRcT clCmSensorEventInformAmf(SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT  *rptentry, SaHpiRdrT *pRdr);
static ClRcT resourceDomainEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT *rptentry );
static ClRcT resourceDomainEventProcess(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT *rptentry );
static ClRcT resourceWatchDogEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT *rptentry );
static ClRcT resourceOEMEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT *rptentry );
static ClRcT resourceUserEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT *rptentry );
static ClRcT resourceSensorEnableEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry);
static ClRcT clCmResourceEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry, SaHpiRdrT *pRdr);
static ClRcT clCmResourceEventInformAmf( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry, SaHpiRdrT *pRdr);
static ClRcT subscribeToPlatformEvents ( SaHpiDomainIdT  *domainId, ClOsalTaskIdT *);
static void clCmFreeEventBufferNode(clCmEventBufferT *node);
static clCmEventBufferT *clCmGetEventBufferNode();
static void clCmRelayHpiEvent(clCmEventBufferT *node);
static clCmEventBufferT *clCmInsertEventBufferNode(clCmEventBufferT *node);
void clCmDeleteBufferedEvents(ClCmEventBufferDomainT *);
void clCmForwardBufferedEvents(ClCmEventBufferDomainT *);
void clCmDomainEntryDel(SaHpiDomainIdT domainId);
void clCmDomainEventBufferEntryDel(SaHpiDomainIdT domainId);
static ClRcT clCmDomainTaskWrapper(void *ptr);
/* Chassis manager context information */
extern ClCmContextT gClCmChassisMgrContext;
extern ClListHeadT gClDomainList; 
extern ClListHeadT gClEventBufferList; 
static ClUint32T gMaxSessionRetries = _OPENHPI_SESSION_RETRIES_DEFAULT;

static ClOsalMutexT gClSensorLock;
 
ClOsalMutexT gClDomainListLock;
extern ClOsalMutexT gClEventBufferLock;
extern uint8_t clCmIsMaster;
ClQueueT   clCmEventQueueHandle = CL_HANDLE_INVALID_VALUE;
extern ClIocNodeAddressT myStandbyNodeAddress;
ClInt8T hpiConnectionStatus;
ClUint32T hpiConnectionCount = 0;

pthread_mutex_t    freeEventQueueMutex = PTHREAD_MUTEX_INITIALIZER;

#define SENSOR_MAP_BITS (12)
#define SENSOR_MAP_SIZE (1 << SENSOR_MAP_BITS )
#define SENSOR_MAP_MASK (SENSOR_MAP_SIZE-1)
#define SENSOR_MAP_HASH(slot) ( (slot) & SENSOR_MAP_MASK )

static struct hashStruct *sensorMap[SENSOR_MAP_SIZE];
typedef struct ClSensorEntry
{
    SaHpiSensorNumT num;
    SaHpiSensorReadingT reading;
    SaHpiEventStateT state;
    ClListHeadT list;
}ClSensorEntryT;

typedef struct ClSensorMap
{
    struct hashStruct hash;
    ClIocNodeAddressT slot;
    ClListHeadT sensorList;
}ClSensorMapT;

clCmEventBufferT  *clCmEventQueueFreeNodeHead = NULL;

ClCmDomainHpiStatusT hpiConnectionstatus[CL_MAX_DOMAINS];
ClUint32T  numberofDomains =0 ;

ClRcT clCmGetDomainConnectionStatusIndex(SaHpiDomainIdT domainId,ClUint32T *index)
{
    ClUint32T i=0;
    for(i =0 ; i < numberofDomains ; i++)
    {
       if(hpiConnectionstatus[i].domainId == domainId)
       {
           *index = i;
           return CL_OK;
       }
    }
    return CL_ERR_DOESNT_EXIST ;
} 

ClRcT  clCmHpiConnectionStatusDisable(SaHpiDomainIdT domainId)
{
    ClRcT rc = CL_OK;
    ClUint32T index = 0;
    
    rc = clCmGetDomainConnectionStatusIndex(domainId, &index);
    if(rc != CL_OK)
    {
       clLog(CL_LOG_SEV_CRITICAL,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED,"Invalid Index return value = [0x%x]",rc);
       return rc;
    }
    if(index < (numberofDomains-1)) 
    {
          memcpy(hpiConnectionstatus+index, hpiConnectionstatus+(numberofDomains-1), sizeof(ClCmDomainHpiStatusT));
    }
    return rc;
} 

/* Called by the Chassis manager to initialize the HPI Interface */
ClRcT clCmHpiInterfaceInitialize(void )
{

    ClRcT rc = CL_OK;
    SaHpiVersionT hpiVersion = 0x0 ; 
    ClCharT *sessionStr = NULL;
    SaHpiDomainInfoT domainInfo;
    ClCmDomainListT *domainListEntry = NULL;
  
    /* Try to find whether the HPI environment is set,if not freak out */

    /* ZSOLT: This is needed only when working with libopenhpi */
    if(getenv("OPENHPI_CONF") == NULL ){
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_INI, "OPENHPI_CONF environment variable is not set. Please set it");
        return CL_CM_ERROR ( CL_ERR_INVALID_PARAMETER );
    }

    hpiVersion = _saHpiVersionGet();
    /*CL_LIST_HEAD_INIT(&gClEventBufferDomainList);*/

    /* Negotiate the Version of HPI */
    if( hpiVersion != SAHPI_INTERFACE_VERSION )
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_INI, "HPI version mismatch between HPI library version [%#x] and CM version [%#x]", hpiVersion, SAHPI_INTERFACE_VERSION);
        return CL_CM_ERROR(CL_ERR_VERSION_MISMATCH);
    }

    if((sessionStr = getenv("CL_ASP_OPENHPI_SESSION_RETRIES")))
    {
        gMaxSessionRetries = atoi(sessionStr);
    }

    if(!gMaxSessionRetries)
        gMaxSessionRetries = _OPENHPI_SESSION_RETRIES_DEFAULT;

    clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_INI,
            "HPI Version found: [%c.0%d.0%d]",
            (unsigned char)(hpiVersion>>16)==0x01?'A':'B',
            (ClUint8T)(hpiVersion>>8),
            (ClUint8T)(hpiVersion));

    rc = clOsalMutexInit(&gClSensorLock);
    CL_ASSERT(rc == CL_OK);
 
    rc = clOsalMutexInit(&gClDomainListLock);
    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_INI," gClDomainListLock initialze return = 0x%x",rc);
    CL_ASSERT( rc == CL_OK);

    domainListEntry = clHeapAllocate( sizeof(ClCmDomainListT));
    if(domainListEntry == NULL )
    {
        clLog(CL_LOG_SEV_ERROR,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," Memory Allocation Failed ");
        return CL_ERR_NO_MEMORY;
    }
    clOsalMutexLock(&gClDomainListLock);
    clListAddTail(&domainListEntry->list, &gClDomainList);
    clOsalMutexUnlock(&gClDomainListLock);

    /* set the execution object handle of Chassis manager */
    if((rc = clEoMyEoObjectSet(gClCmChassisMgrContext.executionObjHandle)) != CL_OK ) {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_INI, "Unable to set execution object handle, error [0x%x]", rc);
        return rc;
    }

    do
    {
        /* Open the Initial session to discover the hardware in the platform*/
        if((rc = openHpiSession (&domainListEntry->sessionId, SAHPI_UNSPECIFIED_DOMAIN_ID))!= CL_OK )
        {
            clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_INI, "Could not open HPI session, error [0x%x]", rc);
        }

    } while (rc != CL_OK);
    hpiConnectionstatus[numberofDomains].hpiConnectionStatus = CL_CM_HPI_CONNECTION_UP;
    hpiConnectionstatus[numberofDomains].hpiConnectionCount++;
    hpiConnectionStatus = CL_CM_HPI_CONNECTION_UP;
    hpiConnectionCount++;
    rc = _saHpiDomainInfoGet(domainListEntry->sessionId,&domainInfo);
    if ( rc != CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," DomainInfoGet failed with rc = [0x%x]",rc);
        _saHpiSessionClose(domainListEntry->sessionId);
        hpiConnectionstatus[numberofDomains].hpiConnectionStatus =  CL_CM_HPI_CONNECTION_DOWN;
        hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
        return rc;
    }
    clLog(CL_LOG_SEV_INFO, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, " Domain Tag = %s Length = %d",domainInfo.DomainTag.Data, domainInfo.DomainTag.DataLength);
    gClCmChassisMgrContext.platformInfo.defaultDomainId = domainInfo.DomainId;
    domainListEntry->domainId = domainInfo.DomainId;
    hpiConnectionstatus[numberofDomains].domainId = domainInfo.DomainId;
    clOsalMutexLock(&gClDomainListLock);
    numberofDomains ++;
    clOsalMutexUnlock(&gClDomainListLock);

    gClCmChassisMgrContext.platformInfo.discoveryInProgress = CL_TRUE;
    rc = clCmDomainThread((void *)&domainInfo.DomainId);
    if(rc != CL_OK)
    {
        clLog(CL_LOG_ERROR, CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," DomainInfoGet failed with rc = [0x%x]",rc);
        /*_saHpiSessionClose(gClCmChassisMgrContext.platformInfo.hpiSessionId);*/
        _saHpiSessionClose(domainListEntry->sessionId);
        hpiConnectionstatus[numberofDomains-1].hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
        hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
    }

    return rc;
}

ClCmDomainListT* clCmDomainEntryGet( SaHpiDomainIdT domainId)
{
    ClListHeadT *iter = NULL;
    clOsalMutexLock(&gClDomainListLock); 
    CL_LIST_FOR_EACH(iter, &gClDomainList)
    {
        ClCmDomainListT *entry = CL_LIST_ENTRY(iter,ClCmDomainListT,list);
        if( entry->domainId == domainId)
        {
            clOsalMutexUnlock(&gClDomainListLock); 
            return entry;
        }
    }
    clOsalMutexUnlock(&gClDomainListLock); 
    return NULL;
}

ClCmDomainListT* clCmDomainInfoGet( SaHpiDomainIdT domainId, ClCmDomainListT *domainInfo)
{
    ClListHeadT *iter = NULL;

    if (!domainInfo)
         return NULL;

    clOsalMutexLock(&gClDomainListLock); 
    CL_LIST_FOR_EACH(iter, &gClDomainList)
    {
        ClCmDomainListT *entry = CL_LIST_ENTRY(iter,ClCmDomainListT,list);
        if( entry->domainId == domainId)
        {
            memcpy(domainInfo, entry, sizeof(ClCmDomainListT));
            clOsalMutexUnlock(&gClDomainListLock); 
            return domainInfo;
        }
    }
    clOsalMutexUnlock(&gClDomainListLock); 
    return NULL;
}
     
void  clCmDomainEntryDel(SaHpiDomainIdT domainId)
{
    ClRcT rc = CL_OK;
    ClListHeadT *current = NULL;
    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CL_LOG_CONTEXT_UNSPECIFIED, " Inside %s ",__FUNCTION__);
    clOsalMutexLock(&gClDomainListLock); 
    CL_LIST_FOR_EACH(current, &gClDomainList)
    {
        ClCmDomainListT *entry = CL_LIST_ENTRY(current,ClCmDomainListT,list);
        if(entry->domainId == domainId)
        { 
            _saHpiSessionClose(entry->sessionId);
            clListDel(current);
            clHeapFree(entry);
            rc = clCmHpiConnectionStatusDisable(domainId);
            if(rc == CL_OK)
            {
                numberofDomains--;
            }
            break;
        }
    }
    clOsalMutexUnlock(&gClDomainListLock); 
    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CL_LOG_CONTEXT_UNSPECIFIED, " Completed %s ",__FUNCTION__);
    return ;
} 
            
static ClRcT clCmDomainThread(void *ptr)
{
    ClRcT rc = CL_OK;
    ClInt16T button;
    button = CL_TRUE;
    ClUint32T  i =0 ;
    SaHpiDomainIdT domainId = *(SaHpiDomainIdT*) ptr;
    ClCmDomainListT *domainEntry = clCmDomainEntryGet(domainId);
    if (domainEntry == NULL)
    {
       domainEntry = clHeapAllocate( sizeof(ClCmDomainListT));
       if( domainEntry == NULL )
       {
           clLog(CL_LOG_SEV_ERROR,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," Memory Allocation Failed ");
           return CL_ERR_NO_MEMORY;
       }
       domainEntry->domainId = domainId;
       clOsalMutexLock(&gClDomainListLock);
       clListAddTail(&domainEntry->list, &gClDomainList);
       clOsalMutexUnlock(&gClDomainListLock);
        
    }
    domainEntry->stopEventProcessing = CL_FALSE;
    rc = openHpiSession( &domainEntry->sessionId, domainEntry->domainId);
    if( rc != CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "Unable to open new HPI session ");
        return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
    }
    rc = clCmGetDomainConnectionStatusIndex(domainId,&i);
    if( rc != CL_OK)
    {
         clLog(CL_LOG_SEV_CRITICAL,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED,"Invalid Index return value = [0x%x]",rc);
         return rc;
    }
    clCmDisplayProgressBar (&button);
    if((rc = DiscoverPlatform(domainEntry->sessionId, domainEntry->domainId, &domainEntry->discoveryInProgress ))!= CL_OK )
    {
        button = CL_FALSE;
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_INI, "DiscoverPlatform failed, error [0x%x]", rc);
        _saHpiSessionClose(domainEntry->sessionId);
        hpiConnectionstatus[i].hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
        hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
        return rc;
    }
    button = CL_FALSE;

    /* Starts the Event listener thread and listens to Platform Events*/
    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_INI, "Launching HPI event receiver loop thread for domain %d ",domainEntry->domainId);
    if ( domainId == gClCmChassisMgrContext.platformInfo.defaultDomainId)
    {
        if((rc = subscribeToPlatformEvents( &domainEntry->domainId , &domainEntry->eventListenerThread ))!= CL_OK )
        {
            clLog(CL_LOG_CRITICAL, AREA_HPI, CTX_INI, "Failed to subscribe to platform events, error [0x%x]", rc);
            _saHpiSessionClose(domainEntry->sessionId);
            hpiConnectionstatus[i].hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
            hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
            return rc;
        }
    }
    else
    {
        hpiEventProcessingLoop((void*)&domainEntry->domainId);
    }
    return CL_OK;
}
ClRcT clCmDomainIdFind( SaHpiDomainIdT domainId)
{
    ClListHeadT *iter = NULL;
    clOsalMutexLock(&gClDomainListLock);
    CL_LIST_FOR_EACH(iter, &gClDomainList)
    {
        ClCmDomainListT *entry = CL_LIST_ENTRY(iter,ClCmDomainListT,list);
        if( entry->domainId == domainId)
        {
            clOsalMutexUnlock(&gClDomainListLock);
            return CL_OK;
        }
    }
    clOsalMutexUnlock(&gClDomainListLock);
    return CL_ERR_INVALID_PARAMETER;
} 

static ClRcT clCmDomainTaskCreate(SaHpiDomainIdT domain_Id)
{
    ClRcT     rc = CL_OK; 
    ClCmDomainListT *domainEntry = NULL;

    clLog( CL_LOG_SEV_DEBUG, " HPI", "INI", "Received DomainId = 0x%x numberofDomains = 0x%x", domain_Id, numberofDomains);
    if( numberofDomains < CL_MAX_DOMAINS ) 
    {

        clLog(CL_LOG_SEV_INFO , " HPI", "INI", " checking whether domain id is unspecified domain id or not ");
        if(( rc = clCmDomainIdFind(domain_Id)) == CL_OK)
        {
            ClCmDomainListT domainInfo;
            memset(&domainInfo,0,sizeof(ClCmDomainListT));
            if(clCmDomainInfoGet(domain_Id,&domainInfo))
            {
                if(!domainInfo.stopEventProcessing)
                {
                    clLog(CL_LOG_SEV_DEBUG,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," Domain Id Already Existed in the List ");
                    return CL_OK;
                }
                else
                {
                    clOsalTaskDelete(domainInfo.eventListenerThread);
                    clCmDomainEventBufferEntryDel(domain_Id);
                    clCmDomainEntryDel(domain_Id);
                }
            }
        }
        domainEntry = clHeapAllocate( sizeof(ClCmDomainListT));
        if( domainEntry == NULL )
        {
            clLog(CL_LOG_SEV_ERROR,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," Memory Allocation Failed ");
            return CL_ERR_NO_MEMORY;
        }
        domainEntry->domainId = domain_Id;
        clOsalMutexLock(&gClDomainListLock);
        clListAddTail(&domainEntry->list, &gClDomainList);
        hpiConnectionstatus[numberofDomains].domainId = domain_Id;
        numberofDomains ++ ;
        clOsalMutexUnlock(&gClDomainListLock);
        
        if( (rc = clOsalTaskCreateAttached( "Perform Domain Related Tasks", 
                        CL_OSAL_SCHED_OTHER, 
                        CL_OSAL_THREAD_PRI_NOT_APPLICABLE, /*Priority*/
                        8*1024*1024, /* Stack Size - Fix to Bug 5266 */ /* The 8MB comes from OpenHPI sample code */
                        (void*(*)(void *))clCmDomainThread,
                        (void *)&domainEntry->domainId,&domainEntry->eventListenerThread))!= CL_OK )

        {
            clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_INI, "Failed to create trhead, error [0x%x]", rc);
        }
    }
    else
    {
        clLog(CL_LOG_SEV_ERROR ,CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," Maximum number of domains supported are [0x%x]",CL_MAX_DOMAINS);
        clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED,CL_LOG_CONTEXT_UNSPECIFIED," Please Configure the MAX domain ");
        rc = CL_ERR_OUT_OF_RANGE;
    }

    return rc;
}
static ClRcT clCmDomainTaskDelete(SaHpiDomainIdT domain_id)
{
    ClCmDomainListT *entry = clCmDomainEntryGet(domain_id);
    if(entry)
    {           
        entry->stopEventProcessing = CL_TRUE;
    }
    return CL_OK;
}
  
void clCmDomainEventBufferEntryDel(SaHpiDomainIdT domainId)
{
    ClListHeadT *current = NULL;
    clLog(CL_LOG_SEV_INFO, AREA_HPI, CL_LOG_CONTEXT_UNSPECIFIED, " Inside %s ",__FUNCTION__);
    clOsalMutexLock(&gClEventBufferLock);
    CL_LIST_FOR_EACH(current, &gClEventBufferList)
    {
        ClCmEventBufferDomainT *entry = CL_LIST_ENTRY(current,ClCmEventBufferDomainT,next);
        if(entry->domainId == domainId)
        {
            clCmDeleteBufferedEvents(entry);
            clListDel(current);
            clHeapFree(entry);
            break;
        }
    }
    clOsalMutexUnlock(&gClEventBufferLock);
    clLog(CL_LOG_SEV_INFO, AREA_HPI, CL_LOG_CONTEXT_UNSPECIFIED, " Completed %s ",__FUNCTION__);
    return ;
}
 
    
/* Called by the Chassis manager to finalize the HPI Interface */
ClRcT clCmHpiInterfaceFinalize(void )
{
    ClRcT rc = CL_OK;
    ClUint32T i = 0;
    ClTimerTimeOutT delay = {1 , 0};

    gClCmChassisMgrContext.platformInfo.stopEventProcessing = CL_TRUE;
    clOsalMutexLock(&gClDomainListLock);
    while(!CL_LIST_HEAD_EMPTY(&gClDomainList))
    {  
       ClCmDomainListT *entry = NULL;
       ClListHeadT *first = gClDomainList.pNext;
       entry = CL_LIST_ENTRY(first,ClCmDomainListT,list);
       (void)clOsalTaskDelay(delay); /* wait a bit before killing the thread */

       clOsalTaskDelete(entry->eventListenerThread);
       _saHpiSessionClose(entry->sessionId);
       clListDel(first);
       clHeapFree(entry);
       
    }
    clOsalMutexUnlock(&gClDomainListLock);
    clOsalMutexLock(&gClEventBufferLock);
    while(!CL_LIST_HEAD_EMPTY(&gClEventBufferList))
    {
        ClListHeadT *first = gClEventBufferList.pNext;
        ClCmEventBufferDomainT *entry = CL_LIST_ENTRY(first,ClCmEventBufferDomainT,next);
        clCmDeleteBufferedEvents(entry);
        clListDel(first);
        clHeapFree(entry);
    }
    clOsalMutexUnlock(&gClEventBufferLock);
        
    /*saHpiSessionClose(gClCmChassisMgrContext.platformInfo.hpiSessionId);*/
    for(i = 0 ; i < numberofDomains ; i++)
         hpiConnectionstatus[i].hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN; 
        
    hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
    sleep(1);

    return rc;
}

static ClRcT scanDomain(SaHpiSessionIdT session_id)
{
    SaHpiEntryIdT    current, next;
    SaHpiDrtEntryT   drt_entry;
    SaErrorT error = SA_OK;
    ClRcT    rc = CL_OK;
    /* list out the domains in the DRT table */
    current = SAHPI_FIRST_ENTRY;
    while( current != SAHPI_LAST_ENTRY )
    {
        if((error = _saHpiDrtEntryGet( session_id, current,  &next,  &drt_entry ))!= SA_OK )
        {
            /* FIXME proper error message comes here */
            if( error == SA_ERR_HPI_NOT_PRESENT && current == SAHPI_FIRST_ENTRY)
            {
                 return CL_OK;
            }
            return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
        }
        /* for each domain entry walk through the RPT and list the RDRs */
        clLog(CL_LOG_SEV_INFO,"HPI","INI"," domain id in the drt table = %d ",drt_entry.DomainId);
        rc =clCmDomainTaskCreate(drt_entry.DomainId); 
        if( rc != CL_OK)
            return rc;
        current = next;
    }
    return CL_OK;
}


static ClRcT clCmCreateResourceHotSwapEvent(SaHpiSessionIdT sessionid, SaHpiRptEntryT *pRptEntry, SaHpiHsStateT  prevHsState,
                                            SaHpiHsStateT  curHsState, ClEventT *HotSwapEvt)
{
    ClRcT rc = CL_OK;
    ClCharT *epath = NULL;
    ClUint32T physical_slot = 0;
    ClUint32T chassisId  = 0;

    if (NULL == HotSwapEvt)
    {
        return CL_ERR_NULL_POINTER;
    }
    /* print the entity path in to the logfile and the previous and the
     *  present states */
    epath = clCmEpath2String(&pRptEntry->ResourceEntity);
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassisId, &physical_slot);
    clLogMultiline(CL_LOG_SEV_NOTICE, AREA_HPI, CTX_EVT,
            "Creating Hot-Swap event for resource [0x%x]:\n"
            "entity:        %s\n"
            "new-state:     %s\n"
            "previous-state:    %s",
            pRptEntry->ResourceId,
            epath,
            oh_lookup_hsstate(curHsState),
            oh_lookup_hsstate(prevHsState));

    HotSwapEvt->PayloadType=SAHPI_ET_HOTSWAP;
    HotSwapEvt->RptEntry = *pRptEntry;
    HotSwapEvt->SlotNumber = physical_slot;
    HotSwapEvt->ChassisId = chassisId;
    HotSwapEvt->Payload.pFruEventPayload.fruId = pRptEntry->ResourceId;
    HotSwapEvt->Payload.pFruEventPayload.physicalSlot = clCmPhysicalSlotFromEntityPath(&(pRptEntry->ResourceEntity));
    HotSwapEvt->Payload.pFruEventPayload.previousState = prevHsState;
    HotSwapEvt->Payload.pFruEventPayload.presentState = curHsState;

    return rc;
}


static ClRcT clCmCreateAndPublishResHsEvt(SaHpiSessionIdT sessionid, SaHpiRptEntryT *pRptEntry, SaHpiHsStateT  prevHsState, SaHpiHsStateT  curHsState)
{
    ClRcT rc = CL_OK;
    ClEventT HotSwapEvt;


    memset(&HotSwapEvt, 0, sizeof(HotSwapEvt));
    clCmCreateResourceHotSwapEvent(sessionid, pRptEntry, prevHsState, curHsState, &HotSwapEvt);

    /* Publish FRU State Change Event to CM Event Channel */
    rc = clCmPublishEvent(CM_FRU_STATE_TRANSITION_EVENT, sizeof(ClEventT), &HotSwapEvt);

    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish FRU HotSwap state event, error[0x%x]", rc);
    }

    return rc;
}
//static DomainInfoGet(
/*Called by Chassis manager to discover the resources in the Platform, This
 * function is called 
 *            i)  During Startup of the Chassis manager . 
 *            ii) Up on User requests from NB Interface. 
 *            iii)Up on any significant events like FRU Arrival,
 *            Departure etc. ZSOLT: really???
 */
static ClRcT DiscoverPlatform(SaHpiSessionIdT  session_id,SaHpiDomainIdT domain_id,ClUint8T *domainDiscoveryInProgress) 
{
    SaErrorT     error = SA_OK ;
    SaHpiDrtEntryT   drt_entry;
    ClRcT        rc = CL_OK;
    
    //gClCmChassisMgrContext.platformInfo.discoveryInProgress = CL_TRUE;
    *domainDiscoveryInProgress= CL_TRUE;
    memset((void*)&drt_entry, 0,sizeof(drt_entry));

    clLog(CL_LOG_SEV_INFO, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Platform discovery started");

    /* discover the domain and all the FRU's in the system */ 
    if((error = _saHpiDiscover( session_id ))!= SA_OK ){
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_DIS, "saHpiDiscover failed: %s", oh_lookup_error(error));
        //gClCmChassisMgrContext.platformInfo.discoveryInProgress = CL_FALSE; 
        *domainDiscoveryInProgress= CL_FALSE;
        return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
    }

    clLog(CL_LOG_SEV_INFO, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "HPI discovery phase complete, starting inventory walk-through");

    /* FIXME: API is not implemented by the plugin ABI uncomment after the API is
     * implemented */ 
#if 0 

    SaHpiEntryIdT    current, next;

    /* list out the domains in the DRT table */ 
    current = SAHPI_FIRST_ENTRY;
    while( current != SAHPI_LAST_ENTRY ){
        if((error = _saHpiDrtEntryGet( session_id, current, &next, &drt_entry ))!= SA_OK ){
            /* FIXME proper error message comes here */
            return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
        }
        /* for each domain entry walk through the RPT and list the RDRs */ 
        WalkThroughDomain( &drt_entry );
        current = next;
    }
#endif 
    
    drt_entry.DomainId = domain_id;
    if ((rc = WalkThroughDomain (&drt_entry))!= CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "WalkThroughDomain failed, error [0x%x]", rc);
    }

    //gClCmChassisMgrContext.platformInfo.discoveryInProgress = CL_FALSE; 
    *domainDiscoveryInProgress= CL_FALSE;

    clLog(CL_LOG_SEV_INFO, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Inventory walk-through completed for Domain [0x%x]", domain_id);

    return CL_OK;
}

static ClSensorMapT *sensorFind(ClIocNodeAddressT slot)
{
    ClUint32T hash = SENSOR_MAP_HASH(slot);
    struct hashStruct *iter;
    for(iter = sensorMap[hash]; iter; iter = iter->pNext)
    {
        ClSensorMapT *map = hashEntry(iter, ClSensorMapT, hash);
        if(map->slot == slot)
            return map;
    }
    return NULL;
}

static ClSensorEntryT *sensorNumFind(ClSensorMapT *map, SaHpiSensorNumT num)
{
    ClListHeadT *iter;
    CL_LIST_FOR_EACH(iter, &map->sensorList)
    {
        ClSensorEntryT *sensor = CL_LIST_ENTRY(iter, ClSensorEntryT, list);
        if(sensor->num == num)
            return sensor;
    }
    return NULL;
}

static __inline__ void sensorLink(ClSensorMapT *map)
{
    ClUint32T hash = SENSOR_MAP_HASH(map->slot);
    hashAdd(sensorMap, hash, &map->hash);
}

static void sensorAdd(SaHpiRdrT *rdr, SaHpiSensorReadingT *reading, 
        SaHpiEventStateT eventState, ClBoolT asserted)
{
    ClSensorMapT *map = NULL;
    ClSensorEntryT *sensor = NULL;
    ClUint32T physical_slot =0 ;
    ClUint32T chassisId = 0;
    ClIocNodeAddressT slot = clCmPhysicalSlotFromEntityPath(&rdr->Entity);
    ClBoolT createFlag = CL_FALSE;
    if(slot == CL_CM_INVALID_PHYSICAL_SLOT)
    {
        clLog(CL_LOG_SEV_TRACE, "SENSOR", "ADD", "clCmPhysicalSlotFromEntityPath() returns a invalid slot number for [%s] entity path",clCmEpath2String(&rdr->Entity));
        return;
    }
    clOsalMutexLock(&gClSensorLock);
    if( !(map = sensorFind(slot)) )
    {
        map = clHeapCalloc(1, sizeof(*map));
        CL_ASSERT(map != NULL);
        map->slot = slot;
        CL_LIST_HEAD_INIT(&map->sensorList);
        sensorLink(map);
        goto addSensor;
    }
    else if( !(sensor = sensorNumFind(map, rdr->RdrTypeUnion.SensorRec.Num) ) )
    {
addSensor:
        sensor = clHeapCalloc(1, sizeof(*sensor));
        CL_ASSERT(sensor != NULL);
        sensor->num = rdr->RdrTypeUnion.SensorRec.Num;
        sensor->state = 0;
        clListAddTail(&sensor->list, &map->sensorList);
        createFlag = CL_TRUE;
    }
    memcpy(&sensor->reading, reading, sizeof(sensor->reading));
    if(asserted)
        sensor->state |= eventState;
    else
        sensor->state &= ~eventState;
    clOsalMutexUnlock(&gClSensorLock);
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&rdr->Entity, &chassisId, &physical_slot);
    clLogNotice("SENSOR", "AUDIT", "Threshold sensor [%d] %s for chassisId [%u] phyicalslot [%u] with event state [%#x] "
            "current event state [%#x]", sensor->num, createFlag ? "added" : "modified",
            chassisId, physical_slot, eventState, sensor->state);
}

static SaHpiSeverityT State2Severity(SaHpiEventStateT in)
{
    if ((in&SAHPI_ES_LOWER_CRIT)||(in&SAHPI_ES_UPPER_CRIT)) return SAHPI_CRITICAL;
    if ((in&SAHPI_ES_LOWER_MAJOR)||(in&SAHPI_ES_UPPER_MAJOR)) return SAHPI_MAJOR;
    if ((in&SAHPI_ES_UPPER_MINOR)||(in&SAHPI_ES_LOWER_MINOR)) return SAHPI_MINOR;
    return SAHPI_OK;
}


static void sensorAudit(SaHpiSessionIdT session, SaHpiResourceIdT resource, SaHpiRdrT *rdr)
{
    SaHpiSensorReadingT    reading = {0};
    SaHpiEventStateT       eventState = 0;
    SaErrorT               rc = SA_OK;
    SaHpiEventT            event;
    SaHpiRptEntryT         rpt;
    //ClRcT                  pluginRc = CL_OK;
    ClRcT                  pluginRc = CL_CM_PLUGIN_OP_PERMITTED | CL_CM_PLUGIN_OP_PUBLISHED;
    SaHpiSensorThresholdsT SensorThresholds;
 
    rc = _saHpiSensorReadingGet( session, resource, rdr->RdrTypeUnion.SensorRec.Num, &reading, &eventState);
    if (rc==SA_OK)
    {
        event.EventDataUnion.SensorEvent.EventState     = eventState;

        /* Prepare a fake event to jam into the user's plugin, so he can filter the event */
        if (eventState)
        {
            if (clPolicyPlugin && clPolicyPlugin->clHpiEventFilter)
            {
                event.Source    = resource;
                event.EventType = SAHPI_ET_SENSOR;
                event.Timestamp = SAHPI_TIME_UNSPECIFIED;
                event.Severity  = State2Severity(eventState);
                event.EventDataUnion.SensorEvent.SensorNum      = rdr->RdrTypeUnion.SensorRec.Num;
                event.EventDataUnion.SensorEvent.SensorType     = rdr->RdrTypeUnion.SensorRec.Type;
                event.EventDataUnion.SensorEvent.Assertion      = 1;
                event.EventDataUnion.SensorEvent.EventState     = eventState;
                event.EventDataUnion.SensorEvent.EventCategory  = rdr->RdrTypeUnion.SensorRec.Category;
                event.EventDataUnion.SensorEvent.TriggerReading = reading;
                switch(event.EventDataUnion.SensorEvent.EventCategory)
                {
                case SAHPI_EC_THRESHOLD:
                    /* Initialize Triggering Threshold to zero */
                    event.EventDataUnion.SensorEvent.TriggerThreshold = reading; 
                    event.EventDataUnion.SensorEvent.TriggerThreshold.Value.SensorInt64 = 0;
                    if(rdr->RdrTypeUnion.SensorRec.ThresholdDefn.IsAccessible)
                    {
                        rc = saHpiSensorThresholdsGet (session, resource, rdr->RdrTypeUnion.SensorRec.Num, &SensorThresholds);
                        switch(event.EventDataUnion.SensorEvent.EventState)
                        {
                        case SAHPI_ES_UPPER_CRIT:
                            event.EventDataUnion.SensorEvent.TriggerThreshold = SensorThresholds.UpCritical;
                            break;
                        case SAHPI_ES_UPPER_MAJOR:
                            event.EventDataUnion.SensorEvent.TriggerThreshold = SensorThresholds.UpMajor;
                            break;
                        case SAHPI_ES_UPPER_MINOR:
                            event.EventDataUnion.SensorEvent.TriggerThreshold = SensorThresholds.UpMinor;
                            break;
                        case SAHPI_ES_LOWER_CRIT:
                            event.EventDataUnion.SensorEvent.TriggerThreshold = SensorThresholds.LowCritical;
                            break;
                        case SAHPI_ES_LOWER_MAJOR:
                            event.EventDataUnion.SensorEvent.TriggerThreshold = SensorThresholds.LowMajor;
                            break;
                        case SAHPI_ES_LOWER_MINOR:
                            event.EventDataUnion.SensorEvent.TriggerThreshold = SensorThresholds.LowMinor;
                            break;
                        default:
                            clLogDebug("SENSOR", "AUDIT", "Invalid event state [%#x]", eventState);
                            break;
                        }
                    }
                    break;
                default:
                            clLogDebug("SENSOR", "AUDIT", "Invalid Event Category [%#x]", event.EventDataUnion.SensorEvent.EventCategory);
                    break; 

                }

                event.EventDataUnion.SensorEvent.OptionalDataPresent = SAHPI_SOD_CURRENT_STATE;
                event.EventDataUnion.SensorEvent.CurrentState   = eventState;
                event.EventDataUnion.SensorEvent.PreviousState  = 0;
                event.EventDataUnion.SensorEvent.Oem            = 0;
                event.EventDataUnion.SensorEvent.SensorSpecific = 0;
                rc = saHpiRptEntryGetByResourceId(session,resource,&rpt);
                if (rc==SA_OK)
                {
                    pluginRc = clPolicyPlugin->clHpiEventFilter(session,&event,rdr,&rpt);
                }
                else
                {
                   clLog(CL_LOG_SEV_ERROR, "SENSOR", "AUDIT", "saHpiRptEntryGetByResourceId failed, Ignoring Event, rc=[0x%x]", rc);
                   pluginRc = CL_CM_PLUGIN_OP_NO_ACTION;
                }
                //if (pluginRc != CL_GET_ERROR_CODE(CL_ERR_OP_NOT_PERMITTED))
                if ( (pluginRc == CL_CM_PLUGIN_OP_NO_ACTION) || (pluginRc == CL_GET_ERROR_CODE(CL_ERR_OP_NOT_PERMITTED)))
                {
                    clLog(CL_LOG_SEV_INFO, "SENSOR", "AUDIT"," Ignoring event");
                }
                else
                { 
                    if ( pluginRc & CL_CM_PLUGIN_OP_PUBLISHED )
                    {
                        if ((rc = clCmSensorEventPublish(session, &event, &rpt, rdr)) != CL_OK)
                        {
                            clLog(CL_LOG_SEV_ERROR, "SENSOR", "AUDIT", "Sensor event publish in Walkthrough failed, error [0x%x]", rc);
                        }
                    }
                    if ( pluginRc & CL_CM_PLUGIN_OP_PERMITTED )
                    {
                        if ((rc = clCmSensorEventInformAmf(session, &event, &rpt, rdr)) != CL_OK)
                        {
                            clLog(CL_LOG_SEV_ERROR, "SENSOR", "AUDIT", "Sensor event notify to amf in Walkthrough failed, error [0x%x]", rc);
                        }
                    }
                  
               }
            }
        }
        //if (pluginRc != CL_GET_ERROR_CODE(CL_ERR_OP_NOT_PERMITTED))
        if (pluginRc & CL_CM_PLUGIN_OP_PERMITTED)
        {
            clLog(CL_LOG_SEV_DEBUG, "SENSOR", "AUDIT", " SensorAudit adding sensor to List");
            sensorAdd(rdr, &reading, event.EventDataUnion.SensorEvent.EventState, CL_TRUE);
        }

    }
}

ClRcT clCmSensorStateGetBySlot(ClIocNodeAddressT slot, SaHpiEventStateT *state)
{
    ClSensorMapT *map;
    ClRcT rc = CL_ERR_NOT_EXIST;
    clOsalMutexLock(&gClSensorLock);
    map = sensorFind(slot);
    if(map)
    {
        ClListHeadT *iter;
        *state = 0;
        CL_LIST_FOR_EACH(iter, &map->sensorList)
        {
            ClSensorEntryT *sensor = CL_LIST_ENTRY(iter, ClSensorEntryT, list);
            
            *state |= sensor->state;
        }
        rc = CL_OK;
    }
    clOsalMutexUnlock(&gClSensorLock);
    return rc;
}

ClRcT clCmSensorStateGetByNumber(ClIocNodeAddressT slot, SaHpiSensorNumT num, SaHpiEventStateT *state)
{
    ClSensorMapT *map;
    ClSensorEntryT *sensor;
    ClRcT rc = CL_ERR_NOT_EXIST;
    clOsalMutexLock(&gClSensorLock);
    if( (map = sensorFind(slot) ) && (sensor = sensorNumFind(map, num)) ) {
        rc = CL_OK;
        *state = sensor->state;
    }
    clOsalMutexUnlock(&gClSensorLock);
    return rc;
}

/*
 * Walk function to walk through the resource and find the RDR's for that
 * resource
 */
ClRcT WalkThroughResource(SaHpiRptEntryT *rpt_entry, SaHpiSessionIdT session_id)
{
    SaHpiResourceIdT    resourceid;
    SaHpiEntryIdT       entryid; 
    SaHpiEntryIdT       nextentryid; 
    SaHpiRdrT        rdr;
    SaErrorT        error = SA_OK ;

    /* Nothing to do if resource does not have RDR capability */
    if ((rpt_entry->ResourceCapabilities & SAHPI_CAPABILITY_RDR) != SAHPI_CAPABILITY_RDR)
    {
        return CL_OK;
    }

    /* walk through the list of RDRs 
     * for this RPT entry and display them
     */
    resourceid = rpt_entry->ResourceId;

    memset(&rdr,0,sizeof(SaHpiRdrT));

    entryid = SAHPI_FIRST_ENTRY;
    while(entryid != SAHPI_LAST_ENTRY )
    {
        if ((error = _saHpiRdrGet(session_id, resourceid, entryid, &nextentryid, &rdr ))!= SA_OK)
        {
            clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "saHpiRdrGet failed: %s", oh_lookup_error(error));
            return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
        }

        switch ( rdr.RdrType )
        {
            case SAHPI_SENSOR_RDR : 
                gClCmChassisMgrContext.platformInfo.no_of_sensors++;
                if((rdr.RdrTypeUnion.SensorRec.Category == SAHPI_EC_THRESHOLD) || (rdr.RdrTypeUnion.SensorRec.ThresholdDefn.IsAccessible == SAHPI_TRUE))
                {
                    sensorAudit(session_id, resourceid, &rdr);
                }
                break;

            case SAHPI_CTRL_RDR:    
                break;

            case SAHPI_INVENTORY_RDR:
                if ((rpt_entry->ResourceCapabilities & SAHPI_CAPABILITY_INVENTORY_DATA) == SAHPI_CAPABILITY_INVENTORY_DATA)
                {
                    gClCmChassisMgrContext.platformInfo.no_of_inventories++;
                }
                break;

            case SAHPI_WATCHDOG_RDR:
                break;

            case SAHPI_ANNUNCIATOR_RDR:
                break;

            default: 
                clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "Unknown resource type detected");
                break;
        }
        entryid = nextentryid;
        memset(&rdr,'\0',sizeof(SaHpiRdrT));

    } /* end while */
    return CL_OK;
}


/* Walk through the RPT table and list all the entries in the RPT Table */ 
ClRcT WalkThroughDomain( SaHpiDrtEntryT   *drt_entry )
{
    SaHpiEntryIdT       current = SAHPI_FIRST_ENTRY, next;
    SaHpiRptEntryT      rpt_entry;
    SaErrorT            error = SA_OK; 
    SaHpiSessionIdT     session_id = 0x0 ; 
    ClRcT               rc = CL_OK;
    SaHpiTimeoutT       autoExtractionTimeOut;
    SaHpiTimeoutT       autoInsertionTimeOut;
    SaHpiHsStateT       hotswapState;

    /* Open the session for the domain id and scan through the RPT Table */ 
    if( openHpiSession( &session_id, drt_entry->DomainId )!= CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "Unable to open new HPI session for resource walk-through");
        return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
    }

    if(drt_entry->DomainId != SAHPI_UNSPECIFIED_DOMAIN_ID)
    {
        _saHpiDiscover(session_id);
    }

    /* Verify that the preset autoinsertion timeout is large enough. */
    if ((error = _saHpiAutoInsertTimeoutGet(session_id, &autoInsertionTimeOut)) != SA_OK)
    {
        clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_DIS, "Could not obtain auto-insert timeout value; proceeding without check");
    }
    else
    {
        clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_DIS, "Current auto-insert timeout is set to [%lld] ms", (long long int) (autoInsertionTimeOut/1000LL));
        /*
         * We want to set this if it is too low (and if it can be altered.
         * If it is set to block, also change it to our preferred value.
         */
        if (autoInsertionTimeOut == SAHPI_TIMEOUT_BLOCK || autoInsertionTimeOut < CL_CM_MIN_AUTOINSERT_TIMEOUT)
        {
            clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_DIS, "Need to adjust auto-insert timeout to [%lld.%09lld] seconds",
                    (long long int) (CL_CM_MIN_AUTOINSERT_TIMEOUT/1000000000LL), (long long int) (CL_CM_MIN_AUTOINSERT_TIMEOUT%1000000000LL));

            if ((error = _saHpiAutoInsertTimeoutSet(session_id, CL_CM_MIN_AUTOINSERT_TIMEOUT))!= SA_OK )
            {
                clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_DIS,
                        "Could not set auto-insert timeout value (%s); " "proceeding nevertheless...", oh_lookup_error(error));
            }
            else
            {
                clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_DIS, "Chassis auto-insert timeout adjusted successfully");
            }
        }
    }

    /* RPT Table is formed now scan each RPT rpt_entry and display the
     * capabilities of the Entities */ 
    memset(&rpt_entry,'\0',sizeof(SaHpiRptEntryT)); 

    current = SAHPI_FIRST_ENTRY;
    while (current != SAHPI_LAST_ENTRY)
    {
        if ((error = _saHpiRptEntryGet(session_id, current, &next, &rpt_entry)) != SA_OK)
        {
            clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "saHpiRPTEntryGet failed: %s", oh_lookup_error(error));
            _saHpiSessionClose(session_id);
            return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
        }
        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_DIS, "HPI resource id [0x%x] is [%s]", rpt_entry.ResourceId, clCmEpath2String(&rpt_entry.ResourceEntity));
        /* find out the FRUs */ 
        if ((rpt_entry.ResourceCapabilities & SAHPI_CAPABILITY_FRU ) && (rpt_entry.ResourceCapabilities & SAHPI_CAPABILITY_MANAGED_HOTSWAP))
        {
            /* Check the auto-extract time first */
            if ((error = _saHpiAutoExtractTimeoutGet(session_id, rpt_entry.ResourceId, &autoExtractionTimeOut)) != SA_OK)
            {
                clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_DIS, "Could not get auto-extract timeout value for resource [0x%x] (%s); proceeding without check",
                        (unsigned int)rpt_entry.ResourceId, oh_lookup_error(error));
            }
            else
            {
                /* Adjust timeout value if needed */
                clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_DIS, "Current auto-extract timeout for [0x%x] is %lld [ms]",
                        rpt_entry.ResourceId, (long long int) (autoExtractionTimeOut/1000LL));

                if (autoExtractionTimeOut == SAHPI_TIMEOUT_BLOCK || autoExtractionTimeOut != CL_CM_MIN_AUTOEXTRACT_TIMEOUT)
                {
                    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_DIS, "Need to adjust auto-extract timeout to [%lld.%09lld] seconds for resource [0x%x]",
                            (long long int)(CL_CM_MIN_AUTOEXTRACT_TIMEOUT/1000000000L), (long long int)(CL_CM_MIN_AUTOEXTRACT_TIMEOUT%1000000000L),
                            rpt_entry.ResourceId);

                    if (rpt_entry.HotSwapCapabilities & SAHPI_HS_CAPABILITY_AUTOEXTRACT_READ_ONLY)
                    {
                        clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_DIS, "Cannot set read-only auto-extract timeout; proceeding...");
                    }
                    else if ((error = _saHpiAutoExtractTimeoutSet(session_id, rpt_entry.ResourceId, CL_CM_MIN_AUTOEXTRACT_TIMEOUT))!= SA_OK )
                    {
                        clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_DIS,
                                "Setting auto-extract timeout for resource [0x%x] failed (%s); proceeding nevertheless...",
                                rpt_entry.ResourceId, oh_lookup_error(error));
                    }
                    else
                    {
                        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_DIS, "Auto-extract timeout successfully adjusted for resource [0x%x]", rpt_entry.ResourceId);
                    }
                }
            }

            error = _saHpiHotSwapStateGet (session_id, rpt_entry.ResourceId,  &hotswapState);
            if (error != SA_OK)
            {
                clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_DIS,
                        "Could not get HotSwap State for resource [0x%x] (%s); " "proceeding without check",
                        (unsigned int)rpt_entry.ResourceId, oh_lookup_error(error));
            }
            else
            {
              clCmCreateAndPublishResHsEvt(session_id, &rpt_entry, hotswapState, hotswapState);
            }

            /* Counting # of FRUs */
            gClCmChassisMgrContext.platformInfo.no_of_frus++; 
        }

        /* display Resource data records assosciated with the resource */
        if ((rc = WalkThroughResource(&rpt_entry, session_id))!= CL_OK)
        {
            clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_DIS, "WalkThroughResource failed, error [0x%x]", rc);
            _saHpiSessionClose(session_id);
            return rc;
        }

        current = next;
        memset(&rpt_entry,'\0',sizeof(SaHpiRptEntryT)); 
    }

    scanDomain(session_id);

    _saHpiSessionClose(session_id);
    return CL_OK;
}


/* Called by the Chassis manager to open a HPI session with the Default Domain
 * Id (defined in SaHpi.h header ). 
 */
static ClRcT openHpiSession( SaHpiSessionIdT *session_id, SaHpiDomainIdT  domain_id )
{
    SaErrorT       error = SA_OK;
    int        retry = 0;
    ClRcT      rc = 1;

    for (retry=1; ((retry<5)&&(rc!=CL_OK)); retry++)
    {
        if ((error = _saHpiSessionOpen(domain_id, session_id, NULL))!= SA_OK)
        {
            clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_INI, "saHpiSessionOpen failed: %s", oh_lookup_error(error));
            rc = CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
            sleep(retry);
        }
        else rc = CL_OK;
    }
    return rc;
}


/*****************************************************************************/
/*                HPI EVENT HANDLING                  */
/*****************************************************************************/

/* Subscribe to the platform events from HPI 
   Platform Events like :
   i) FRU Arrival, Departure etc.
   ii) Sensor asserts and Deasserts etc.
 */
static ClRcT subscribeToPlatformEvents(SaHpiDomainIdT *domainId, ClOsalTaskIdT *pEventListenerThread)
{
    ClRcT rc = CL_OK;

    /* Create the Event Listener Thread and process the Events */
    if ((rc = clOsalTaskCreateAttached(
                    "Event listener thread",
                    CL_OSAL_SCHED_OTHER,
                    CL_OSAL_THREAD_PRI_NOT_APPLICABLE, /*Priority*/
                    8*1024*1024, /* Stack Size - Fix to Bug 5266 */
                    /* The 8MB comes from OpenHPI sample code */
                    (void*(*)(void *))hpiEventProcessingLoop,
                    (void *)domainId,
                    pEventListenerThread
                    ))!= CL_OK )
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_INI, "Failed to allocate trhead, error [0x%x]", rc);
    }

    return rc;
}

static char * eventEpathToString(SaHpiEventT *event, SaHpiEntityPathT *epath)
{
    guint id;
    static oh_big_textbuffer buffer;
    SaHpiEntityPathT ep;
    SaErrorT err = SA_OK;

    id = event->Source;
    if (epath) {
        err  = oh_decode_entitypath(epath, &buffer);
    } else {
        err = oh_entity_path_lookup(id, &ep);
        if (err) {
            clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Could not determine entity path for resource id [0x%x]", id);
        }
        else {
            err  = oh_decode_entitypath(&ep, &buffer);
        }
    }
    if (err == SA_OK)
        return (char*)&buffer.Data;
    else
        return NULL;

}

static char * eventDetailsToString(SaHpiEventT *event, SaHpiEntityPathT *epath)
{
    ClCharT str[SAHPI_MAX_TEXT_BUFFER_LENGTH], *str2;
    static oh_big_textbuffer buffer;
    SaHpiTextBufferT tmpbuffer;
    SaErrorT err = SA_OK;

    (void)oh_init_bigtext(&buffer);

    /* type */
    snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH, "\ntype:       %s", oh_lookup_eventtype(event->EventType));
    oh_append_bigtext(&buffer, str);

    /* severity */
    snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH, "\nseverity:       %s", oh_lookup_severity(event->Severity));
    oh_append_bigtext(&buffer, str);

    /* epath */
    str2 = eventEpathToString(event, epath);
    if (str2 != NULL)
    {
        snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH, "\nentity:     %s", str2);
        oh_append_bigtext(&buffer, str);
    }

    /* resource id */
    snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH, "\nresid:      %d (0x%x)", event->Source, event->Source);
    oh_append_bigtext(&buffer, str);

    switch (event->EventType)
    {
        case SAHPI_ET_SENSOR:

            snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH,
                    "\nsensor-id:      %d (0x%x)", event->EventDataUnion.SensorEvent.SensorNum, event->EventDataUnion.SensorEvent.SensorNum);
            oh_append_bigtext(&buffer, str);

            snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH,
                    "\nsensor-type:    %s", oh_lookup_sensortype(event->EventDataUnion.SensorEvent.SensorType));
            oh_append_bigtext(&buffer, str);

            snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH, "\nasserted:       %s", (event->EventDataUnion.SensorEvent.Assertion == SAHPI_TRUE) ?  "TRUE" : "FALSE");
            oh_append_bigtext(&buffer, str);

            err = oh_decode_eventstate(event->EventDataUnion.SensorEvent.EventState, event->EventDataUnion.SensorEvent.EventCategory, &tmpbuffer);
            snprintf(str, SAHPI_MAX_TEXT_BUFFER_LENGTH, "\nstate:      %s", err ? "<unknown>" : (char*)tmpbuffer.Data);
            oh_append_bigtext(&buffer, str);

            break;
        case SAHPI_ET_HOTSWAP:
            break;
        default:
            break;
    }

    /* we do assume here that the data is an ASCII null-terminated string */
    return (char*)buffer.Data;
}

/* Event listener loop listens for the Events from the HPI */
static ClRcT hpiEventProcessingLoop(void *ptr)
{
#define SAHPI_EVENT_TIMEOUT (5 * 1000 * 1000 * 1000LL) 
    SaHpiRptEntryT    rptentry ; 
    SaHpiRdrT         rdr;
    SaHpiEventT       event;
    SaErrorT          error = SA_OK ;
    ClUint32T         errcnt = 0;
    ClRcT             rc = CL_OK;
    SaHpiDomainIdT    domainId = 0x0;
    SaHpiSessionIdT   sessionid = 0x0;
    ClBoolT           sessionidValid = CL_FALSE;
    SaHpiEvtQueueStatusT  qstatus; 
    ClUint8T          printSeverity;
    ClUint32T         i = 0;
    ClCmEventBufferDomainT  *eventBuffer = NULL ;
    ClCmDomainListT         *domainInfo = NULL;
   
    memset(&rptentry, 0, sizeof(rptentry));
    memset(&rdr, 0, sizeof(rdr));
    memset(&event, 0, sizeof(event));
    domainId = *(SaHpiDomainIdT*)ptr;

    domainInfo = clCmDomainEntryGet(domainId);
    if(domainInfo == NULL)
    {
        clLog(CL_LOG_SEV_DEBUG, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, " Invalid Domain ID Received Existing from hpiEventProcessing Loop");     
        return CL_CM_ERROR(CL_ERR_NOT_EXIST); 
    } 
    clLog(CL_LOG_SEV_DEBUG, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Received Domain Id = %d",domainId);
    
    clCmGetDomainConnectionStatusIndex(domainId,&i);
   
    do
    {
        eventBuffer = clHeapAllocate(sizeof(ClCmEventBufferDomainT));
    }while( eventBuffer == NULL );
    eventBuffer->domainId = domainId;
    eventBuffer->clCmEventQueueHead = NULL;
    eventBuffer->clCmEventQueueTail = NULL;
    eventBuffer->clCmEventQueueNodeCount = 0;
    eventBuffer->peerHpiLastCount = CL_CM_HPI_CONNECTION_VAL; /* For first time we don't know the peerhpiconnection count*/
    clOsalMutexLock(&gClEventBufferLock); 
    clListAddTail(&eventBuffer->next,&gClEventBufferList);
    clOsalMutexUnlock(&gClEventBufferLock); 
         

    clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_INI, "Entering HPI event receive loop");
    clLog(CL_LOG_SEV_TRACE,AREA_HPI, CTX_INI, " HPI event receive loop for domain = %d",domainId);

    while(!gClCmChassisMgrContext.platformInfo.stopEventProcessing && !domainInfo->stopEventProcessing )
    {
        /* Don't loop forever waiting; fail over! */
        errcnt++;  /* errcnt will be set to 0 if we ever receive an event */
        if (errcnt > gMaxSessionRetries)
        {
            clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Cannot connect to HPI events within a reasonable time period.  Aborting to cause failover.");
            exit(1); /* For multiple Domains We need to evaluate this*/
        }

        /* If we don't have a session, then call hpi event subscribe to subscribe for platform events */
        if (!sessionidValid)
        {
            if ((error = openHpiSession (&sessionid, domainId))!= CL_OK )
            {
                clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_INI, "openHPI session reopen failed.");
                hpiConnectionstatus[i].hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
                hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
                sleep(5);
                continue;
            }
            if ((error = _saHpiSubscribe( sessionid )) != SA_OK  )
            {
                _saHpiSessionClose(sessionid);
                hpiConnectionstatus[i].hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
                hpiConnectionStatus = CL_CM_HPI_CONNECTION_DOWN;
                sessionid = 0;
                clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_INI, "saHpiSubscribe failed: %s. Retrying...", oh_lookup_error(error));
                sleep(5);
                continue;
            }
            hpiConnectionstatus[i].hpiConnectionCount++;
            hpiConnectionstatus[i].hpiConnectionStatus = CL_CM_HPI_CONNECTION_UP;
            hpiConnectionStatus = CL_CM_HPI_CONNECTION_UP;
            clLog(CL_LOG_SEV_DEBUG,CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, " DomainId = %d hpiConnectionStatus = %d hpiConnectionCount =%d",
                            hpiConnectionstatus[i].domainId,hpiConnectionstatus[i].hpiConnectionStatus, hpiConnectionstatus[i].hpiConnectionCount);
            sessionidValid = CL_TRUE;
        }

        /* Get the events from the Domain Event Table */
        error = _saHpiEventGet(sessionid, SAHPI_EVENT_TIMEOUT, &event, &rdr, &rptentry, &qstatus);
        if(error != SA_OK)
        {
            /* No events came in; no big deal */
            if(error == SA_ERR_HPI_TIMEOUT)
            {
                errcnt=0;
                continue; 
            }

            sleep(5);        

            clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_EVT, "saHpiEventGet failed: %s", oh_lookup_error(error));
            _saHpiUnsubscribe( sessionid );
            _saHpiSessionClose( sessionid );
            sessionid = 0;
            sessionidValid = CL_FALSE;        
            /* the rest of the code processes the event.  So do not run it. */
            continue;

        }

        /* We got a good event, so reset the sequential error count */ 
        errcnt = 0;


        /* Check whether we lost any events due to queue filling */
        if( qstatus == SAHPI_EVT_QUEUE_OVERFLOW )
        {
            clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_EVT, "HPI event queue overflow (some events may have been lost)");
        }
        
        ClRcT pluginOps = CL_CM_PLUGIN_OP_PERMITTED | CL_CM_PLUGIN_OP_PUBLISHED;
        if (clPolicyPlugin && clPolicyPlugin->clHpiEventFilter)
        {
            clLogMultiline(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT, "(Prefiltered) HPI event received:%s", eventDetailsToString(&event, &(rdr.Entity)));
            pluginOps = clPolicyPlugin->clHpiEventFilter(sessionid,&event,&rdr,&rptentry);
            clLog(CL_LOG_SEV_INFO,"plu","ops","pluginOps =[ ox%x]",pluginOps);
            if((pluginOps == CL_CM_PLUGIN_OP_NO_ACTION) || (pluginOps == CL_GET_ERROR_CODE(CL_ERR_OP_NOT_PERMITTED)))
            {
                continue;
            }   
        }
         
        if (CL_CM_FALSE == clCmIsMaster)
        {
           clCmEventBufferT *bufferEventNode = NULL;
           bufferEventNode = clCmGetEventBufferNode();
           bufferEventNode->sessionId = sessionid;
           bufferEventNode->domainId = domainId;
           bufferEventNode->pluginOps = pluginOps;
           bufferEventNode->domainQueueInfo = &eventBuffer;
           memcpy(&(bufferEventNode->event), &event, sizeof(event));
           memcpy(&(bufferEventNode->rptentry), &rptentry, sizeof(rptentry));
           memcpy(&(bufferEventNode->rdr), &rdr, sizeof(rdr));
           
           if( NULL ==  clCmInsertEventBufferNode(bufferEventNode))
           {
               clLogMultiline(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Error While buffering the event:%s, ErrCode:[0x%x]", eventDetailsToString(&event, &(rdr.Entity)), rc);
           }
           else
           {
               clLogMultiline(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT, "Successfully Buffered the Event:%s", eventDetailsToString(&event, &(rdr.Entity)));
               
           }
           
           continue;

        }
        else
        { 
            /* Standby CM become Active, Flush the standby Domain Queue before publishing it.*/
            if (0 < eventBuffer->clCmEventQueueNodeCount)
            {
                clCmForwardBufferedEvents(eventBuffer);
            }
        }
        /*clLogMultiline(CL_LOG_DEBUG, AREA_HPI, CTX_EVT, "My Standby Node Address:[0x%x]",myStandbyNodeAddress); */
        /* only the active system contoller card supposed to generate events */
        if(event.Severity == SAHPI_CRITICAL)
        {    
            printSeverity = CL_LOG_SEV_EMERGENCY;
        }
        else if (event.Severity == SAHPI_MAJOR)
        {    
            printSeverity = CL_LOG_SEV_ALERT;
        }
        else
        {   
             printSeverity = CL_LOG_SEV_INFO; 
        }
        clLogMultiline(printSeverity, AREA_HPI, CTX_EVT, "HPI event received:%s", eventDetailsToString(&event, &(rdr.Entity)));
        // Have only active CM process events in case of openhpid or CM failure & restart on one node
        switch(event.EventType)
        {

            case  SAHPI_ET_RESOURCE:
				 /*Process the Resource event */ 
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = clCmResourceEventPublish(sessionid, &event, &rptentry, &rdr))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource event publish failed, error [0x%x]", rc);
                    }
                }
                if ( pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
                {
                    if ((rc = clCmResourceEventInformAmf(sessionid, &event, &rptentry, &rdr))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource event notification  failed, error [0x%x]", rc);
                    }
                }
                continue;

            case  SAHPI_ET_HOTSWAP :
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = resourceHotSwapEventPublish(sessionid, &event, &rptentry))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource Hotswap event publish failed, error [0x%x]", rc);
                    }
                }
                if ( pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
                {
                    if ((rc = resourceHotSwapEventInformAmf(sessionid, &event, &rptentry))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource Hotswap event notification failed, error [0x%x]", rc);
                    }
                }

                 /*Process the Hotswap event */
                 #if 0 
                 if ((rc = resourceHotSwapEventProcess(sessionid, &event,&rptentry))!= CL_OK )
                 {
                 clLog(CL_LOG_ERROR, AREA_HPI, CTX_EVT, "Hotswap event processing failed, error [0x%x]", rc);
                 }
                 #endif
                 continue;   

            case  SAHPI_ET_SENSOR  :
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = clCmSensorEventPublish(sessionid, &event, &rptentry, &rdr))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Sensor event publish failed, error [0x%x]", rc);
                    }
                }
                if ( pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
                {
                    if ((rc = clCmSensorEventInformAmf(sessionid, &event, &rptentry, &rdr))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Sensor  event notification failed, error [0x%x]", rc);
                    }
                }

                #if 0
                if ((rc = clCmSensorEventProcess(sessionid, &event, &rptentry, &rdr)) != CL_OK)
                {
                    clLog(CL_LOG_ERROR, AREA_HPI, CTX_EVT, "Sensor event processing failed, error [0x%x]", rc);
                }
                #endif
#ifndef HPI_CHANNEL_EVENT_SIM
                continue; 
#endif
            case  SAHPI_ET_DOMAIN:
                clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT,"Received Domain event for Domain Id [%d] on the Domain Id [%d]", event.EventDataUnion.DomainEvent.DomainId, domainId);
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
          	    if ((rc = resourceDomainEventPublish(sessionid, &event, &rptentry))!= CL_OK )
                    {
                        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT,"Domain event publish failed, error [0x%x]", rc);
                    }
                }
                if ( pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
                {
          	    if ((rc = resourceDomainEventProcess(sessionid, &event, &rptentry))!= CL_OK )
                    {
                         clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Domain  event process failed, error [0x%x]", rc);
                    }
                }
#ifndef HPI_CHANNEL_EVENT_SIM
                continue;
#endif
            case  SAHPI_ET_WATCHDOG:
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = resourceWatchDogEventPublish(sessionid, &event, &rptentry))!= CL_OK )
                    {
                        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "WatchDog event publish failed, error [0x%x]", rc);
                    }
                }
#ifndef HPI_CHANNEL_EVENT_SIM
                continue;
#endif
            case  SAHPI_ET_OEM:
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = resourceOEMEventPublish(sessionid, &event, &rptentry))!= CL_OK )
                    {
                        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Hotswap event publish failed, error [0x%x]", rc);
                    }
                }
#ifndef HPI_CHANNEL_EVENT_SIM
                continue;
#endif
            case  SAHPI_ET_USER:
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = resourceUserEventPublish(sessionid, &event, &rptentry))!= CL_OK )
                    {
                        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Hotswap event publish failed, error [0x%x]", rc);
                    }
                }
#ifndef HPI_CHANNEL_EVENT_SIM
                continue;
#endif
            case  SAHPI_ET_SENSOR_ENABLE_CHANGE:     
                if ( pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
                {
                    if ((rc = resourceSensorEnableEventPublish(sessionid, &event, &rptentry))!= CL_OK )
                    {
                        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Hotswap event publish failed, error [0x%x]", rc);
                    }
                }

                continue;

            default :
                clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Invalid event type [%d] received, ignoring", event.EventType);
                continue;

        }
    } /*end while Loop */
    
    clLog(CL_LOG_SEV_INFO, AREA_HPI, CL_LOG_CONTEXT_UNSPECIFIED, "Leaving HPI event receive loop for Domain %d",domainId);

    /* Avoiding crash in case of both terminate callback and domain removal event occur simultaneously*/
    if(!gClCmChassisMgrContext.platformInfo.stopEventProcessing && (domainInfo != NULL) && domainInfo->stopEventProcessing)
    {
       clCmDomainEventBufferEntryDel(domainId);
       clCmDomainEntryDel(domainId);
    }
    
    if(sessionidValid)
    {
        _saHpiUnsubscribe( sessionid );
        _saHpiSessionClose( sessionid );
    }
    /*_saHpiSessionClose(gClCmChassisMgrContext.platformInfo.hpiSessionId);*/
    return CL_OK;

#undef SAHPI_EVENT_TIMEOUT
}
/* ET Resource event processing :
*/
static ClRcT clCmResourceEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry, SaHpiRdrT *pRdr)
{

    ClRcT rc = CL_OK;
    ClEventT * resEvt;
    ClUint32T physical_slot = 0;
    ClUint32T chassis_id = 0 ;

    /* Publish FRU State Change Event to HPI Event Channel */
    resEvt = clHeapAllocate(sizeof(ClEventT));
    if (NULL == resEvt)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to allocate memory for resource event, error[0x%x]", rc);
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassis_id, &physical_slot);
    resEvt->PayloadType = SAHPI_ET_RESOURCE;
    resEvt->RptEntry = *pRptEntry;
    resEvt->SlotNumber = physical_slot; 
    resEvt->ChassisId = chassis_id; 
    resEvt->Payload.pHpiResEventPayload.ResourceEventType = pEvent->EventDataUnion.ResourceEvent.ResourceEventType;

    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT), resEvt);

    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish resource event, error[0x%x]", rc);
    }

    clHeapFree(resEvt);
    return rc;
}

static ClRcT clCmResourceEventInformAmf( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry, SaHpiRdrT *pRdr) 
{

    ClRcT rc = CL_OK;
    if ((pEvent->EventDataUnion.ResourceEvent.ResourceEventType == SAHPI_RESE_RESOURCE_FAILURE)
                        && (pEvent->Severity == SAHPI_MAJOR))
    {
        clLogMultiline(CL_LOG_SEV_NOTICE, AREA_HPI, CTX_EVT,
                            "MAJOR HPI resource failure reported on [%s]\n"
                            "Treating it as surprise extraction",
                            eventEpathToString(pEvent, &(pRdr->Entity)));

        rc = clCmHotSwapCpmUpdate(pRptEntry, CL_CM_BLADE_SURPRISE_EXTRACTION);
        if (CL_OK != rc)
        {
            clLog(CL_LOG_SEV_NOTICE, AREA_HPI, CTX_EVT, "Failed to report surprise extraction, error [0x%x]", rc);
                        /* Do not return as the info in cor needs to be updated*/
        }
    }
    return rc;
}




/* Resource hotswap processing :
   -----------------------------
   Resources can follow either fullhotswap state machine or simplified hotswap
   state machine. FRUs which follow full hotswap state machine mostly will be
   running ASP instance on them and they have a physical slot in their entity
   paths, so the rest of the system needs to be notified about the node
   departure. FRUs which follow simplified hotswap model doesnt need any
   notification to the rest of the system since they will be FANS,
   POWERSUPPLIES we need to just update the management view of the Domain.

   This assumption is not true, as BCT blades also follow Simple Hot Swap
   Model.
 */
static ClRcT resourceHotSwapEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry)
{
    ClRcT rc = CL_OK;
    ClCharT *epath = NULL;
    ClEventT * HotSwapEvtResource;
    ClUint32T physical_slot = 0;
    ClUint32T chassisId     = 0;

    /* print the entity path in to the logfile and the previous and the
     *  present states */
    epath = clCmEpath2String(&pRptEntry->ResourceEntity);
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassisId, &physical_slot);
    clLogMultiline(CL_LOG_SEV_NOTICE, AREA_HPI, CTX_EVT,
            "Hot-swap event received for resource [0x%x]:\n"
            "entity:        %s\n"
            "new-state:     %s\n"
            "previous-state:    %s",
            pEvent->Source,
            epath,
            oh_lookup_hsstate(pEvent->EventDataUnion.HotSwapEvent.HotSwapState),
            oh_lookup_hsstate(pEvent-> EventDataUnion.HotSwapEvent.PreviousHotSwapState));

    /* Publish FRU State Change Event to CM Event Channel */
    HotSwapEvtResource = clHeapAllocate(sizeof(ClEventT));
    if (NULL == HotSwapEvtResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    HotSwapEvtResource->PayloadType=SAHPI_ET_HOTSWAP;
    HotSwapEvtResource->RptEntry = *pRptEntry;
    HotSwapEvtResource->SlotNumber = physical_slot;
    HotSwapEvtResource->ChassisId = chassisId;
    HotSwapEvtResource->Payload.pFruEventPayload.fruId = pEvent->Source;
    HotSwapEvtResource->Payload.pFruEventPayload.physicalSlot = clCmPhysicalSlotFromEntityPath(&(pRptEntry->ResourceEntity));
    HotSwapEvtResource->Payload.pFruEventPayload.previousState = pEvent->EventDataUnion.HotSwapEvent.PreviousHotSwapState;
    HotSwapEvtResource->Payload.pFruEventPayload.presentState = pEvent->EventDataUnion.HotSwapEvent.HotSwapState;
    rc = clCmPublishEvent(CM_FRU_STATE_TRANSITION_EVENT, sizeof(ClEventT), HotSwapEvtResource);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish FRU state transition event, error[0x%x]", rc);
    }

    clHeapFree(HotSwapEvtResource);
    return rc;

}
static ClRcT resourceHotSwapEventInformAmf( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry)
{
    ClRcT rc = CL_OK;
    /* Check the previous state of the FRU whether its a valid transition or
     *  Invalid Transition. If its a valid transition then query CPMG whether
     *  we can should allow the arrival or departure of the FRU in to the
     *  system. This query is an asynchronous query and the FRU waits in the
     *  EXTRACTION_PENDING state or INSERTION_PENDING state until the CPMG
     *  responds back to the Chassis manager query 
     */
    if((pRptEntry->ResourceCapabilities & SAHPI_CAPABILITY_MANAGED_HOTSWAP) ==  SAHPI_CAPABILITY_MANAGED_HOTSWAP)
    {

        rc = clCmHandleFruHotSwapEvent( sessionid, pEvent->EventDataUnion.HotSwapEvent.PreviousHotSwapState, pEvent->EventDataUnion.HotSwapEvent.HotSwapState, pRptEntry);
    }
    else
    {
        /*
         * We cannot control the hotswap of the resource so we just update the
         * management view
         */
        rc = clCmSimpleHotSwapHandler(sessionid, pEvent, pRptEntry);
    }
    return rc;

}

#if 0
static ClRcT resourceHotSwapEventProcess( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry) 
{ 
    ClRcT rc = CL_OK;
    ClCharT *epath = NULL;
    ClEventT * HotSwapEvtResource;

    /* print the entity path in to the logfile and the previous and the
     *  present states */
    epath = clCmEpath2String(&pRptEntry->ResourceEntity);

    clLogMultiline(CL_LOG_SEV_NOTICE, AREA_HPI, CTX_EVT,
            "Hot-swap event received for resource [0x%x]:\n"
            "entity:        %s\n"
            "new-state:     %s\n"
            "previous-state:    %s",
            pEvent->Source,
            epath,
            oh_lookup_hsstate(pEvent->EventDataUnion.HotSwapEvent.HotSwapState),
            oh_lookup_hsstate(pEvent-> EventDataUnion.HotSwapEvent.PreviousHotSwapState));

    /* Publish FRU State Change Event to CM Event Channel */
    HotSwapEvtResource = clHeapAllocate(sizeof(ClEventT));
    if (NULL == HotSwapEvtResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    HotSwapEvtResource->PayloadType=SAHPI_ET_HOTSWAP;
    HotSwapEvtResource->RptEntry = *pRptEntry;
    HotSwapEvtResource->Payload.pFruEventPayload.fruId = pEvent->Source;
    HotSwapEvtResource->Payload.pFruEventPayload.physicalSlot = clCmPhysicalSlotFromEntityPath(&(pRptEntry->ResourceEntity));
    HotSwapEvtResource->Payload.pFruEventPayload.previousState = pEvent->EventDataUnion.HotSwapEvent.PreviousHotSwapState;
    HotSwapEvtResource->Payload.pFruEventPayload.presentState = pEvent->EventDataUnion.HotSwapEvent.HotSwapState;

    rc = clCmPublishEvent(CM_FRU_STATE_TRANSITION_EVENT, sizeof(ClEventT), HotSwapEvtResource);

    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish FRU state transition event, error[0x%x]", rc);
    }

    clHeapFree(HotSwapEvtResource);

    /* Check the previous state of the FRU whether its a valid transition or
     *  Invalid Transition. If its a valid transition then query CPMG whether
     *  we can should allow the arrival or departure of the FRU in to the
     *  system. This query is an asynchronous query and the FRU waits in the
     *  EXTRACTION_PENDING state or INSERTION_PENDING state until the CPMG
     *  responds back to the Chassis manager query 
     */
    if((pRptEntry->ResourceCapabilities & SAHPI_CAPABILITY_MANAGED_HOTSWAP) ==  SAHPI_CAPABILITY_MANAGED_HOTSWAP)
    {

        rc = clCmHandleFruHotSwapEvent( sessionid, pEvent->EventDataUnion.HotSwapEvent.PreviousHotSwapState, pEvent->EventDataUnion.HotSwapEvent.HotSwapState, pRptEntry);
    }
    else
    {
        /*
         * We cannot control the hotswap of the resource so we just update the
         * management view
         */
        rc = clCmSimpleHotSwapHandler(sessionid, pEvent, pRptEntry);
    }
    return rc;
}
#endif



/* 
   Description
   ------------
   Handles the response from the CPMG for the previous query made. 
 */
    void 
clCmHPICpmResponseHandle(ClCpmCmMsgT * pCpm2CmMsg)
{
    SaErrorT saError;
    SaHpiRptEntryT rptEntry;

    clLogMultiline(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_REQ,
            "Response from AMF arrived:\n"
            "message-type:      %d\n"
            "physical-slot:     %d\n"
            "resource-id:       %d (0x%x)\n"
            "response-type:     %d",
            pCpm2CmMsg->cmCpmMsg.cmCpmMsgType,
            pCpm2CmMsg->cmCpmMsg.physicalSlot,
            pCpm2CmMsg->cmCpmMsg.resourceId,
            pCpm2CmMsg->cmCpmMsg.resourceId,
            pCpm2CmMsg->cpmCmMsgType);
    /* sessionId Needs to be corrected for Multiple Domains */
    saError = _saHpiRptEntryGetByResourceId( gClCmChassisMgrContext.platformInfo.hpiSessionId, pCpm2CmMsg->cmCpmMsg.resourceId, &rptEntry);
    if (SA_OK !=saError)
    {
        clLog(CL_LOG_SEV_WARNING,  AREA_HPI, CTX_REQ, "saHpiRptEntryGetByResourceId failed: %s", oh_lookup_error(saError));
        return ;
    }

    switch(pCpm2CmMsg->cmCpmMsg.cmCpmMsgType)
    {
        case CL_CM_BLADE_REQ_EXTRACTION:
            {
                /* No-op for now */
                break ;
            }
        case CL_CM_BLADE_REQ_INSERTION:
            {
                /* No-op for now */
                break;
            }
        case CL_CM_BLADE_SURPRISE_EXTRACTION:
            {
                /* No Handling from CM */
                break;

            }
        default :
            {
                clLog(CL_LOG_SEV_ERROR,  AREA_HPI, CTX_REQ, "Invalid message type [%d] from AMF", pCpm2CmMsg->cmCpmMsg.cmCpmMsgType);
            }
    }
    return ;
}

ClRcT clCmHotSwapCpmUpdate(SaHpiRptEntryT *pRptEntry, ClCmCpmMsgTypeT cm2CpmMsgType)
{
    ClRcT rc = CL_OK;
    ClUint32T physicalSlot = CL_CM_INVALID_PHYSICAL_SLOT;
    ClCmCpmMsgT cm2CpmMsg;
    ClCharT *epath = NULL;

    epath = clCmEpath2String(&pRptEntry->ResourceEntity);
    if (NULL == epath)
        epath = "<could not decode>";

    clLog(CL_LOG_SEV_ALERT, AREA_HPI, CTX_EVT, "Hot-swap state change event received for [%s]", epath);

    physicalSlot = clCmPhysicalSlotFromEntityPath(&(pRptEntry->ResourceEntity));
    if (CL_CM_INVALID_PHYSICAL_SLOT != physicalSlot)
    {
        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Entity path resolves to slot [%d], informing AMF", physicalSlot);
    }
    else 
    {
        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Could not resolve entity path to a physical slot, ignoring surprise extraction");
        return CL_CM_ERROR( CL_ERR_CM_HPI_INVALID_PHY_SLOT);
    }

    cm2CpmMsg.physicalSlot=physicalSlot;
    cm2CpmMsg.cmCpmMsgType=cm2CpmMsgType;
    /**FIXME AMC handling, fill up the subslot sending 0 as of now */
    cm2CpmMsg.subSlot = 0;
    cm2CpmMsg.resourceId=pRptEntry->ResourceId;

    rc = clCpmHotSwapEventHandle(&cm2CpmMsg);
    if (rc!=CL_OK ) 
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Failed to inform AMF, error [0x%x]", rc);
    }
    return rc;
}



/*
 * Simply update the information in COR so that the management sees the
 * updated view of the domain.
 */
ClRcT clCmSimpleHotSwapHandler (SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT  * pRptEntry)
{
    ClRcT rc = CL_OK;

    if (SAHPI_HS_STATE_ACTIVE == pEvent->EventDataUnion.HotSwapEvent.HotSwapState) 
    {
        /* No Information needs to be sent to CPM-G*/
    }
    else if (SAHPI_HS_STATE_NOT_PRESENT == pEvent->EventDataUnion.HotSwapEvent.HotSwapState) 
    {
        /* Inform to CPM-G */
        rc = clCmHotSwapCpmUpdate(pRptEntry, CL_CM_BLADE_SURPRISE_EXTRACTION);
    }
    else
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Unexpected state [%u] for simple FRU", pEvent->EventDataUnion.HotSwapEvent.HotSwapState);
        return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }
    return rc ;
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*             HPI Event HANDLING                     */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/*NOTE:
  -----
  HPI Provides a DAT (DomainAlarmTable ) which maintains the list of alarms in
  the system . HPI is responsible for clearing and appending the alarms to the
  DAT.DAT interface suits best for SNMP kind of apps which walk through the
  Table. 
  Event with severity = major | minor | critical can be regarded as alarms and
  we publish the events with those severities to ASP as alarms. 
 */
extern ClCmEventCorrelatorT clCmEventCorrelationConfigTable[]; 
extern ClCmUserPolicyT cmUserPolicy[];

ClRcT clCmUserPolicyExecute(SaHpiEntityPathT  *pEntity, SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT  *pRptEntry, SaHpiRdrT *pRdr)
{
    ClRcT rc=CL_OK;
#if 0    // Deprecated in favor of a plugin architecture
    ClUint32T index=0, found=0;
    ClCharT *epath;
    epath = clCmEpath2String(pEntity);
    if (NULL == epath)
        epath = "<could not decode>";

    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT, "Searching for user policy handler for [%s]", epath);
    while (SAHPI_ENT_UNSPECIFIED != 
            cmUserPolicy[index].impactedEntity.Entry[0].EntityType)
    {
        if (SAHPI_TRUE==oh_cmp_ep(pEntity,
                    &(cmUserPolicy[index].impactedEntity)))
        {
            if (NULL!= cmUserPolicy[index].userEventHandler)
            {
                clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT,
                        "Calling user policy handler for [%s]", epath);
                rc=cmUserPolicy[index].userEventHandler(sessionid,
                        pEvent,
                        pRptEntry,
                        pRdr);
                found = 1;
                break;                    
            }
        }
        index++;
    }
#endif    
    /* Deprecating the out-of-band user policy handler, in preference to the .so so not going to log about it */
    /*  if (found == 0) clLog(CL_LOG_DEBUG, AREA_HPI, CTX_EVT, "No user policy found for [%s]", epath); */
    return rc;
}

ClUint32T clCmNumImpactedEntitiesGet( int index, SaHpiEntityPathT  *pReportingEntity)
{

    ClUint32T j=0;
#if 0    
    if(NULL!=clCmEventCorrelationConfigTable[index].pImpactedEntities)
    {
        while (SAHPI_ENT_UNSPECIFIED != 
                clCmEventCorrelationConfigTable[index].pImpactedEntities[j].Entry[0].EntityType)

            j++;

    }
#endif    
    return j;
}

ClRcT clCmImpactedEntitiesGet(SaHpiEntityPathT  *pReportingEntity, SaHpiEntityPathT  **ppImpactedEntities, ClUint32T *pNumImpactedEntities)
{
#if 0 // Deprecated in favor of the plugin
    ClUint32T i=0;
    ClUint32T bFound =CL_FALSE;
    while (SAHPI_ENT_UNSPECIFIED != 
            clCmEventCorrelationConfigTable[i].eventReporter.Entry[0].EntityType)
    {
        if(SAHPI_TRUE==oh_cmp_ep(pReportingEntity, 
                    &(clCmEventCorrelationConfigTable[i].eventReporter)))

        {
            bFound =CL_TRUE;
            *pNumImpactedEntities = clCmNumImpactedEntitiesGet(i, pReportingEntity);
            *ppImpactedEntities=(SaHpiEntityPathT *)clHeapAllocate(sizeof(SaHpiEntityPathT)*
                    (*pNumImpactedEntities));
            if (NULL==(*ppImpactedEntities))                            
            {
                return CL_ERR_NO_MEMORY;
            }
            memcpy(*ppImpactedEntities, 
                    clCmEventCorrelationConfigTable[i].pImpactedEntities,
                    *pNumImpactedEntities);
            break ; 
        }
        i++;
    }
    if (CL_FALSE==bFound)
    {
    }
#endif    
    return CL_OK;
}


static ClRcT clCmNSIEventHandle( SaHpiSessionIdT sessionId, SaHpiEventT *pEvent, SaHpiRptEntryT  *pRptEntry, SaHpiRdrT *pRdr)
{
    /* Get the list of Impacted Entities, other than itself */
    SaHpiEntityPathT  *pImpactedEntities=NULL;   
    ClUint32T numEntities=0; 
    ClUint32T counter=0; 
    ClRcT rc;
    ClCharT *epath;

    epath = clCmEpath2String(&pRdr->Entity);
    if (NULL == epath)
        epath = "<could not decode>";

    clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT, "Non-service impacting event received on entity [%s]", epath);

    /* SaHpiTextBufferT *epath = NULL; */
    rc = clCmImpactedEntitiesGet(&(pRdr->Entity), &pImpactedEntities, &numEntities);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Cannot handle NSI event due to memory allocation error, error [0x%x]", rc);
        return rc;
    }
    /* call user defined handler(s) */
    rc = clCmUserPolicyExecute(&(pRdr->Entity), sessionId, pEvent, pRptEntry, pRdr);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Custom NSI event handler failed, error [0x%x] for [%s]", rc, eventEpathToString(pEvent, &pRdr->Entity));
    }

    if (pImpactedEntities!=NULL)
    {
        clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT, "Found [%d] indirectly impacted entities, calling their handlers...", numEntities);
        for (counter=0; counter < numEntities; counter++)
        {
            rc = clCmUserPolicyExecute(&(pImpactedEntities[counter]), sessionId, pEvent, pRptEntry, pRdr);
            if (CL_OK != rc)
            {
                clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Custom NSI event handler failed, error [0x%x] for [%s]",
                        rc, eventEpathToString(pEvent, &(pImpactedEntities[counter])));
            }
        }
        clHeapFree(pImpactedEntities);
    }
    return CL_OK;
}

void clCmInformAMFBySlot(ClUint32T slotId, ClBoolT asserted, SaHpiTimeT timeStamp)
{
    ClRcT rc = CL_OK;
    ClCmCpmMsgT cmCpmMsg;
    const ClCharT *msg;

    memset(&cmCpmMsg, 0, sizeof(ClCmCpmMsgT));

    cmCpmMsg.physicalSlot = slotId;

    if (asserted)
    {
        cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_NODE_ERROR_REPORT;
        msg = "(imminent) failure";
    }
    else
    {
        cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_NODE_ERROR_CLEAR;
        msg = "recovery";
    }
    rc = clCpmHotSwapEventHandle(&cmCpmMsg);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Failed to notify AMF about [%s] of slot [%d]", msg, slotId);
        return;
    }
    clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "CM successfully notified AMF on [%s] of slot [%d]", msg, slotId);
    return;
}

void clCmInformAMFByEpath(SaHpiEntityPathT *pEpath, ClBoolT asserted, SaHpiTimeT timeStamp)
{
    ClUint32T slotId = CL_CM_INVALID_PHYSICAL_SLOT ;
    ClCharT *epath_str = NULL;

    epath_str = clCmEpath2String(pEpath);
    if (NULL == epath_str)
        epath_str = "<could not decode>";

    slotId = clCmPhysicalSlotFromEntityPath(pEpath);
    if (CL_CM_INVALID_PHYSICAL_SLOT != slotId)
    {
        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Entity [%s] resolves to slot [%d], informing AMF on service impacting event", epath_str, slotId);
        clCmInformAMFBySlot(slotId, asserted, timeStamp);
    }
    else 
    {
        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Could not resolve entity path [%s] to a physical slot, ignoring", epath_str);
        /* do nothing  as the entity path might not have physical slot in it's
         * entity path*/
    }
}

static ClRcT clCmSIEventHandle( SaHpiSessionIdT sessionid, SaHpiEventT *pEvent, SaHpiRptEntryT  *pRptentry, SaHpiRdrT *pRdr)
{
    SaHpiEntityPathT  *pImpactedEntities=NULL;   
    ClUint32T numEntities=0; 
    ClUint32T counter=0; 
    ClRcT rc=CL_OK;
    ClCharT *epath;
    ClBoolT event_asserted=0;
    SaErrorT error = SA_OK;
    SaHpiSensorReadingT reading;
    SaHpiEventStateT eventState;

    /*
       SaHpiTextBufferT *epath = NULL;
     */ 

    epath = clCmEpath2String(&pRdr->Entity);
    if (NULL == epath)
        epath = "<could not decode>";

    clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Service impacting event received on entity [%s]", epath);

    /* We are here only if this was a sensor event type */
    assert(pEvent->EventType == SAHPI_ET_SENSOR);
    event_asserted = pEvent->EventDataUnion.SensorEvent.Assertion == SAHPI_TRUE;
    clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "This indicates the %s of the event", event_asserted ? "assertion" : "deassertion");

    if(pEvent->EventDataUnion.SensorEvent.EventCategory == SAHPI_EC_THRESHOLD)
    {
        sensorAdd(pRdr, &pEvent->EventDataUnion.SensorEvent.TriggerReading, pEvent->EventDataUnion.SensorEvent.EventState, event_asserted);
    }
    /*
     * In case of deassertion of threshold sensor type, we read the sensor
     * state and only deassert if the both MAJOR and CRITICIAL state is cleared.
     */
    if (!event_asserted && pEvent->EventDataUnion.SensorEvent.EventCategory == SAHPI_EC_THRESHOLD)
    {
        if ((error = _saHpiSensorReadingGet( sessionid, pRptentry->ResourceId, pEvent->EventDataUnion.SensorEvent.SensorNum, &reading, &eventState))!= SA_OK )
        {
            clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_EVT, "Cannot read sensor state; using what is indicated in event");
            /* we proceed based on current event_asserted value */
        }
        else 
        {
            if ((eventState & (SAHPI_ES_LOWER_MAJOR | SAHPI_ES_LOWER_CRIT | SAHPI_ES_UPPER_MAJOR | SAHPI_ES_UPPER_CRIT)) != 0)
            {
                clLog(CL_LOG_SEV_WARNING, AREA_HPI, CTX_EVT, "Deassertion event ignored because sensor is still in MAJOR or CRITICAL state");
                /* We bail out as sensor is still in asserted state */
                return rc;
            }
        } 
    }

    /* Report directly impacted FRUs to AMF */
    clCmInformAMFByEpath(&(pRdr->Entity), event_asserted, pEvent->Timestamp);

    /*
     * Report on indirectly impacted FRUs if any. This is a customizable list
     * under <model>/config
     */
    rc = clCmImpactedEntitiesGet(&(pRdr->Entity), &pImpactedEntities, &numEntities);
    if (CL_OK!=rc)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Cannot derive indirectly impacted entities due to failed memory alloc, error [0x%x]", rc);
        return rc;
    }
    if (pImpactedEntities!=NULL)
    {
        clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT, "Found [%d] indirectly impacted entities, calling their handlers...", numEntities);
        for (counter=0;counter < numEntities; counter++)
        {
            clCmInformAMFByEpath(&(pImpactedEntities[counter]), event_asserted, pEvent->Timestamp);
        }
        clHeapFree(pImpactedEntities);
    }
    return CL_OK;
}

static ClRcT clCmSensorEventPublish(SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT  *pRptentry, SaHpiRdrT *pRdr)
{
    
    ClRcT       rc = CL_OK;
    ClCmSensorEventInfoT *pSensorEventPayload;
    ClUint32T   physical_slot = 0;
    ClUint32T   chassis_id = 0;
    /*To get the node Name */

    /*Defining a new wrapper datatype as per Andrew's Comment */
    ClEventT *SensorResource;
    SensorResource = clHeapAllocate(sizeof(ClEventT)+sizeof(ClCmSensorEventInfoT));
    if (NULL == SensorResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptentry->ResourceEntity, &chassis_id, &physical_slot);
    pSensorEventPayload = (ClCmSensorEventInfoT *) SensorResource->Payload.pAlarmPayload.buff;

    pSensorEventPayload->rptEntry = *pRptentry;
    pSensorEventPayload->rdr = *pRdr;
    pSensorEventPayload->sensorEvent = pEvent->EventDataUnion.SensorEvent;

    SensorResource->PayloadType=SAHPI_ET_SENSOR;
    SensorResource->RptEntry=*pRptentry;
#ifdef HPI_CHANNEL_EVENT_SIM
    clNameSet(&(SensorResource->NodeName),(const char *)"ScNode0");
#endif
    SensorResource->SlotNumber = physical_slot;
    SensorResource->ChassisId = chassis_id;
    SensorResource->Payload.pAlarmPayload.alarmState = ((pEvent->EventDataUnion.SensorEvent.Assertion == SAHPI_TRUE) ?  CL_ALARM_STATE_ASSERT : CL_ALARM_STATE_CLEAR);
    SensorResource->Payload.pAlarmPayload.category = CL_ALARM_CATEGORY_EQUIPMENT;
    SensorResource->Payload.pAlarmPayload.probCause = CL_ALARM_PROB_CAUSE_EQUIPMENT_MALFUNCTION;
    SensorResource->Payload.pAlarmPayload.specificProblem = 0;
    SensorResource->Payload.pAlarmPayload.len = sizeof(ClCmSensorEventInfoT);

    rc = clCmPublishEvent(CM_ALARM_EVENT,sizeof(ClEventT)+sizeof(ClCmSensorEventInfoT),SensorResource);

    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish sensor event, error[0x%x]", rc);
    }
    /* Publish Event on HPI CHANNEL also */
    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT)+sizeof(ClCmSensorEventInfoT), SensorResource);

    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish sensor event, error[0x%x]", rc);
    }
    
    clHeapFree(SensorResource);
    return rc;
}

static ClRcT clCmSensorEventInformAmf(SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT  *rptentry, SaHpiRdrT *pRdr)
{
    ClRcT rc = CL_OK;
  
    clLog(CL_LOG_SEV_INFO,"SEN","PROCESS","%s",__FUNCTION__);
    if (pEvent->Severity > SAHPI_MAJOR)
    {
        rc = clCmNSIEventHandle(sessionid, pEvent, rptentry, pRdr);

    }
    /* Service Impacting Events Handling */
    else
    {
        rc = clCmSIEventHandle(sessionid, pEvent, rptentry, pRdr);

    }
    return rc;  

}

/*Name : resourceDomainEventPublish
 *Note : The Code needs to be customized  
 *       since we don't have much hardware info 
 * static ClRcT clCmSensorEventProcess( SaHpiSessionIdT sessionId,
 SaHpiEventT *pEvent,
 SaHpiRptEntryT *pRptentry,
 SaHpiRdrT *pRdr)
 */

static ClRcT resourceDomainEventPublish( SaHpiSessionIdT sessionId, SaHpiEventT *pEvent, SaHpiRptEntryT *pRptEntry)
{
    ClRcT     rc = CL_OK;
    ClEventT *DomainResource;
    ClUint32T physical_slot = 0;
    ClUint32T chassis_id = 0;
    DomainResource = clHeapAllocate(sizeof(ClEventT));

    if (NULL == DomainResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassis_id, &physical_slot);
#ifdef HPI_CHANNEL_EVENT_SIM
    pEvent->EventType = SAHPI_ET_DOMAIN;
    pEvent->EventDataUnion.DomainEvent.Type = SAHPI_DOMAIN_REF_ADDED;
    pEvent->EventDataUnion.DomainEvent.DomainId = SAHPI_UNSPECIFIED_DOMAIN_ID;
    clNameSet(&(DomainResource->NodeName),(const char *)"ScNode0");
#endif
    DomainResource->RptEntry=*pRptEntry;/* New Addition */
    DomainResource->SlotNumber = physical_slot;
    DomainResource->ChassisId= chassis_id;
    DomainResource->PayloadType = SAHPI_ET_DOMAIN;
    DomainResource->Payload.pHpiDomainPayload.Type = pEvent->EventDataUnion.DomainEvent.Type;
    DomainResource->Payload.pHpiDomainPayload.DomainId = pEvent->EventDataUnion.DomainEvent.DomainId;

    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT),DomainResource);    
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish domain event, error[0x%x]", rc);
    }
    else
    {
        clLog(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Published domain event successfully ");
    }

    clHeapFree(DomainResource);
    return rc;

}//End of resourceDomainEventPublish
static ClRcT resourceDomainEventProcess( SaHpiSessionIdT sessionId, SaHpiEventT *pEvent, SaHpiRptEntryT *pRptEntry)
{

    ClRcT     rc = CL_OK;
    if(pEvent->EventDataUnion.DomainEvent.Type == SAHPI_DOMAIN_REF_ADDED)
    {
        if(clCmIsMaster)
        {
            rc = clCmDomainTaskCreate(pEvent->EventDataUnion.DomainEvent.DomainId);
        }
        else
        {
            SaHpiDomainIdT *domainIdPtr = NULL;
            domainIdPtr = (SaHpiDomainIdT *)clHeapAllocate(sizeof(SaHpiDomainIdT));
            if(domainIdPtr)
            {
                *domainIdPtr = pEvent->EventDataUnion.DomainEvent.DomainId;
                rc = clOsalTaskCreateDetached("clCmDomainTaskWrapperforStandby",CL_OSAL_SCHED_RR,100,0,(void *(*)(void *))clCmDomainTaskWrapper,(void *)domainIdPtr);
                if(rc != CL_OK)
                {
                    clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Creation of a thread for processing domain Added events failed with rc [0x%x]", rc);
                }
            }
            else
            {
                clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED, "Memory Allocation is failed");
                rc = CL_CM_ERROR(CL_ERR_NO_MEMORY);
            }
        }
    }
    else
    {
        rc = clCmDomainTaskDelete(pEvent->EventDataUnion.DomainEvent.DomainId);

    }
    return rc;
}

static ClRcT clCmDomainTaskWrapper(void *domainIdptr)
{
   ClRcT rc = CL_OK; 
   SaHpiDomainIdT    domainId = *(SaHpiDomainIdT*)domainIdptr;
   rc = clCmDomainTaskCreate(domainId);
   clHeapFree(domainIdptr);
   return rc;
} 

/* Name : resourceWatchDogEventPublish
 * Note : The Code needs to be customized
 *    since we don't have much hardware info
 */
static ClRcT resourceWatchDogEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry)
{
    ClRcT     rc = CL_OK;
    ClEventT *WatchDogResource;
    ClUint32T physical_slot = 0;
    ClUint32T chassis_id = 0;
    WatchDogResource = clHeapAllocate(sizeof(ClEventT));

    if (NULL == WatchDogResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassis_id, &physical_slot);
#ifdef HPI_CHANNEL_EVENT_SIM
    pEvent->EventDataUnion.WatchdogEvent.WatchdogNum = 1;
    pEvent->EventDataUnion.WatchdogEvent.WatchdogAction = SAHPI_WAE_RESET;
    pEvent->EventDataUnion.WatchdogEvent.WatchdogPreTimerAction = SAHPI_WPI_MESSAGE_INTERRUPT;
    pEvent->EventDataUnion.WatchdogEvent.WatchdogUse = SAHPI_WTU_OS_LOAD;
    clNameSet(&(WatchDogResource->NodeName),(const char *)"ScNode0");
#endif
    WatchDogResource->RptEntry=*pRptEntry;/* New Addition */
    WatchDogResource->SlotNumber = physical_slot;
    WatchDogResource->ChassisId = chassis_id;
    WatchDogResource->PayloadType = SAHPI_ET_WATCHDOG;
    WatchDogResource->Payload.pHpiWatchdogPayload.WatchdogNum = pEvent->EventDataUnion.WatchdogEvent.WatchdogNum;
    WatchDogResource->Payload.pHpiWatchdogPayload.WatchdogAction = pEvent->EventDataUnion.WatchdogEvent.WatchdogAction;
    WatchDogResource->Payload.pHpiWatchdogPayload.WatchdogPreTimerAction = pEvent->EventDataUnion.WatchdogEvent.WatchdogPreTimerAction;
    WatchDogResource->Payload.pHpiWatchdogPayload.WatchdogUse = pEvent->EventDataUnion.WatchdogEvent.WatchdogUse;

    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT),WatchDogResource);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish WatchDog Payload event to HPI Channel, error[0x%x]", rc);
        return rc;
    }
    clHeapFree(WatchDogResource);
    return rc;

}//End of resourceWatchDog Event publish

/* Name : resourceOEMEventPublish
 * Note : The Code needs to be customized
 *    since we don't have much hardware info
 */
static ClRcT resourceOEMEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry)
{
    ClRcT     rc = CL_OK;
    ClEventT *OemResource;
    ClUint32T physical_slot = 0;
    ClUint32T chassis_id = 0;

    OemResource = clHeapAllocate(sizeof(ClEventT));
    if (NULL == OemResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassis_id, &physical_slot);
#ifdef HPI_CHANNEL_EVENT_SIM
    pEvent->EventDataUnion.OemEvent.OemEventData.DataType = SAHPI_TL_TYPE_BINARY;
    pEvent->EventDataUnion.OemEvent.OemEventData.Language = SAHPI_LANG_BENGALI;
    pEvent->EventDataUnion.OemEvent.OemEventData.DataLength = 30;
    clNameSet(&(OemResource->NodeName),(const char *)"ScNode0");
#endif
    OemResource->RptEntry=*pRptEntry;  /* New Resource */
    OemResource->SlotNumber = physical_slot;  
    OemResource->ChassisId = chassis_id;  
    OemResource->PayloadType = SAHPI_ET_OEM;
    OemResource->Payload.pHpiOemEventPayload.MId=0;
    OemResource->Payload.pHpiOemEventPayload.OemEventData.DataType = pEvent->EventDataUnion.OemEvent.OemEventData.DataType;
    OemResource->Payload.pHpiOemEventPayload.OemEventData.Language = pEvent->EventDataUnion.OemEvent.OemEventData.Language;
    OemResource->Payload.pHpiOemEventPayload.OemEventData.DataLength = pEvent->EventDataUnion.OemEvent.OemEventData.DataLength;

    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT),OemResource);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish Oem Payload event to HPI Channel, error[0x%x]", rc);
        return rc;
    }
    clHeapFree(OemResource);
    return rc;
}//End of resourceOEMEventPublish


/* Name : resourceUserEventPublish
 * Note : The Code needs to be customized
 *    since we don't have much hardware info
 */
static ClRcT resourceUserEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry)
{

    ClRcT     rc = CL_OK;
    ClEventT *UserEvtResource;
    ClUint32T physical_slot = 0;
    ClUint32T chassis_id = 0;

    UserEvtResource = clHeapAllocate(sizeof(ClEventT));
    if (NULL == UserEvtResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassis_id, &physical_slot);
#ifdef HPI_CHANNEL_EVENT_SIM
    pEvent->EventDataUnion.ResourceEvent.ResourceEventType = SAHPI_RESE_RESOURCE_ADDED;
    clNameSet(&(UserEvtResource->NodeName),(const char *)"ScNode0");
#endif
    UserEvtResource->RptEntry=*pRptEntry;
    UserEvtResource->SlotNumber = physical_slot; 
    UserEvtResource->ChassisId = chassis_id;
    UserEvtResource->PayloadType = SAHPI_ET_USER;
    UserEvtResource->Payload.pHpiUserEventPayload.UserEventData.DataType = pEvent->EventDataUnion.UserEvent.UserEventData.DataType;
    UserEvtResource->Payload.pHpiUserEventPayload.UserEventData.Language = pEvent->EventDataUnion.UserEvent.UserEventData.Language;
    UserEvtResource->Payload.pHpiUserEventPayload.UserEventData.DataLength = pEvent->EventDataUnion.UserEvent.UserEventData.DataLength;
    memcpy(UserEvtResource->Payload.pHpiUserEventPayload.UserEventData.Data, pEvent->EventDataUnion.UserEvent.UserEventData.Data, 
               pEvent->EventDataUnion.UserEvent.UserEventData.DataLength);

    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT),UserEvtResource);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish User event Payload event to HPI Channel, error[0x%x]", rc);
        return rc;
    }
    clHeapFree(UserEvtResource);
    return rc;

}//End of resourceUserEventPublish


static ClRcT  resourceSensorEnableEventPublish( SaHpiSessionIdT sessionid, SaHpiEventT * pEvent, SaHpiRptEntryT * pRptEntry)
{
    ClRcT     rc = CL_OK;
    ClEventT * SensorEnableResource;
    ClUint32T physical_slot = 0;
    ClUint32T chassis_id = 0;

    SensorEnableResource = clHeapAllocate(sizeof(ClEventT));
    if (NULL == SensorEnableResource)
    {
        return CL_ERR_NULL_POINTER;
    }
    clCmGetPhysicalSlotandChassisIdFromEntityPath(&pRptEntry->ResourceEntity, &chassis_id, &physical_slot);
#ifdef HPI_CHANNEL_EVENT_SIM
    pEvent->EventDataUnion.SensorEnableChangeEvent.SensorNum = 8 ;
    pEvent->EventDataUnion.SensorEnableChangeEvent.SensorType = SAHPI_TEMPERATURE;
    pEvent->EventDataUnion.SensorEnableChangeEvent.EventCategory = SAHPI_EC_SEVERITY;
    pEvent->EventDataUnion.SensorEnableChangeEvent.SensorEnable = 1;
    pEvent->EventDataUnion.SensorEnableChangeEvent.AssertEventMask = SAHPI_ES_LOWER_CRIT;
    pEvent->EventDataUnion.SensorEnableChangeEvent.DeassertEventMask = SAHPI_ES_UPPER_MAJOR;
    pEvent->EventDataUnion.SensorEnableChangeEvent.CurrentState = SAHPI_ES_UPPER_MINOR;
    clNameSet(&(SensorEnableResource->NodeName),(const char *)"ScNode0");
#endif
    SensorEnableResource->RptEntry=*pRptEntry;
    SensorEnableResource->SlotNumber = physical_slot;
    SensorEnableResource->ChassisId = chassis_id;
    SensorEnableResource->PayloadType = SAHPI_ET_SENSOR_ENABLE_CHANGE;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.SensorNum = pEvent->EventDataUnion.SensorEnableChangeEvent.SensorNum;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.SensorType = pEvent->EventDataUnion.SensorEnableChangeEvent.SensorType;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.EventCategory = pEvent->EventDataUnion.SensorEnableChangeEvent.EventCategory;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.SensorEnable = pEvent->EventDataUnion.SensorEnableChangeEvent.SensorEnable;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.AssertEventMask = pEvent->EventDataUnion.SensorEnableChangeEvent.AssertEventMask;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.DeassertEventMask= pEvent->EventDataUnion.SensorEnableChangeEvent.DeassertEventMask;
    SensorEnableResource->Payload.pHpiSensorEnableChangePayload.CurrentState = pEvent->EventDataUnion.SensorEnableChangeEvent.CurrentState;

    rc = clCmHpiPublishEvent(CM_ALARM_EVENT, sizeof(ClEventT),SensorEnableResource);
    if (CL_OK != rc)
    {
        clLog(CL_LOG_SEV_CRITICAL, AREA_HPI, CTX_EVT, "Failed to publish SensorEnableChange Payload event to HPI Channel, error[0x%x]", rc);
        return rc;
    }
    clHeapFree(SensorEnableResource);
    return rc;

}//End of resourceSensorEnableEventPublish


static clCmEventBufferT *clCmInsertEventBufferNode(clCmEventBufferT *node)
{
    ClCmEventBufferDomainT *eventDomainQueue = *(node->domainQueueInfo);
    clOsalMutexLock(&eventDomainQueue->eventQueueMutex);
    if(eventDomainQueue->clCmEventQueueTail)
    {
        node->next = NULL;
        eventDomainQueue->clCmEventQueueTail->next = node;
        eventDomainQueue->clCmEventQueueTail = node; 
    }
    else
    {
        eventDomainQueue->clCmEventQueueTail = eventDomainQueue->clCmEventQueueHead = node;
    }
    eventDomainQueue->clCmEventQueueNodeCount = eventDomainQueue->clCmEventQueueNodeCount + 1;
    clOsalMutexUnlock(&eventDomainQueue->eventQueueMutex);
    return eventDomainQueue->clCmEventQueueHead; 
}
#if 0
static clCmEventBufferT *clCmDeleteEventBufferNode()
{
    clCmEventBufferT *deletedNode = NULL;
    clOsalMutexLock(&eventDomainQueue->eventBufferMutex);
    /*pthread_mutex_lock(&eventBufferMutex);*/
    if (eventDomainQueue->clCmEventQueueHead)
    {
        deletedNode = eventDomainQueue->clCmEventQueueHead;
        eventDomainQueue->clCmEventQueueHead = eventDomainQueue->clCmEventQueueHead->next;
        deletedNode->next = NULL;
        eventDomainQueue->clCmEventQueueNodeCount = eventDomainQueue->clCmEventQueueNodeCount - 1;
    }
    if(NULL == eventDomainQueue->clCmEventQueueHead)
    {
       eventDomainQueue->clCmEventQueueTail = NULL;
    }
    clOsalMutexUnlock(&eventDomainQueue->eventBufferMutex);
    /*pthread_mutex_unlock(&eventBufferMutex);*/
    return deletedNode;
}
#endif

void clCmForwardBufferedEvents(ClCmEventBufferDomainT *eventDomainQueue)
{
    clCmEventBufferT *deletedNode = NULL;
          
    clOsalMutexLock(&eventDomainQueue->eventQueueMutex);
    while(eventDomainQueue->clCmEventQueueHead)
    {
        if (eventDomainQueue->clCmEventQueueHead)
        {
            deletedNode = eventDomainQueue->clCmEventQueueHead;
            eventDomainQueue->clCmEventQueueHead = eventDomainQueue->clCmEventQueueHead->next;
            deletedNode->next = NULL;
            eventDomainQueue->clCmEventQueueNodeCount = eventDomainQueue->clCmEventQueueNodeCount - 1;
        }
        if(NULL == eventDomainQueue->clCmEventQueueHead)
        {
            eventDomainQueue->clCmEventQueueTail = NULL;
        }
        clCmRelayHpiEvent(deletedNode);
        clCmFreeEventBufferNode(deletedNode);
    }
    clOsalMutexUnlock(&eventDomainQueue->eventQueueMutex);
    return ;
}

void clCmDeleteBufferedEvents(ClCmEventBufferDomainT *eventDomainQueue)
{
    clCmEventBufferT *deletedNode = NULL;
    
    clOsalMutexLock(&eventDomainQueue->eventQueueMutex);
    while(eventDomainQueue->clCmEventQueueHead)
    {
        if (eventDomainQueue->clCmEventQueueHead)
        {
            deletedNode = eventDomainQueue->clCmEventQueueHead;
            eventDomainQueue->clCmEventQueueHead = eventDomainQueue->clCmEventQueueHead->next;
            deletedNode->next = NULL;
            eventDomainQueue->clCmEventQueueNodeCount = eventDomainQueue->clCmEventQueueNodeCount - 1;
        }
        if(NULL == eventDomainQueue->clCmEventQueueHead)
        {
            eventDomainQueue->clCmEventQueueTail = NULL;
        }
        if( deletedNode->event.EventType == SAHPI_ET_DOMAIN )
        {
            /* Standby will perform same action as active except publishing in to the channel */
            deletedNode->pluginOps &= ~CL_CM_PLUGIN_OP_PUBLISHED;
            clCmRelayHpiEvent(deletedNode);
              
        } 
        clCmFreeEventBufferNode(deletedNode);
    }
    clOsalMutexUnlock(&eventDomainQueue->eventQueueMutex);
    return ; 
}

static clCmEventBufferT *clCmGetEventBufferNode()
{
    clCmEventBufferT  *node = NULL;
    uint8_t           index = 0;

    pthread_mutex_lock(&freeEventQueueMutex);
 
    if (NULL == clCmEventQueueFreeNodeHead) {
        clCmEventQueueFreeNodeHead  = (clCmEventBufferT *) malloc(100 * sizeof(clCmEventBufferT));
        for(index = 0; index < 99; index++) {
            clCmEventQueueFreeNodeHead[index].next = (clCmEventQueueFreeNodeHead + index + 1); 
        }
        clCmEventQueueFreeNodeHead[index].next = NULL;
    }
    node = clCmEventQueueFreeNodeHead;
    clCmEventQueueFreeNodeHead = clCmEventQueueFreeNodeHead->next;
    node->next = NULL;
    pthread_mutex_unlock(&freeEventQueueMutex);

    return node;
}

static void clCmFreeEventBufferNode(clCmEventBufferT *node)
{
    if (node)
    {
        pthread_mutex_lock(&freeEventQueueMutex);
        node->next = clCmEventQueueFreeNodeHead;
        clCmEventQueueFreeNodeHead = node;
        pthread_mutex_unlock(&freeEventQueueMutex);
    }
}

static void clCmRelayHpiEvent(clCmEventBufferT *node)
{
    ClRcT rc = CL_OK;

    clLogMultiline(CL_LOG_SEV_INFO, AREA_HPI, CTX_EVT, "Relay HPI event received:%s", eventDetailsToString(&node->event, &(node->rdr.Entity)));
    // Have both standby and active CM process events in case of openhpid or CM failure & restart on one node
    switch(node->event.EventType)
    {

        case  SAHPI_ET_RESOURCE:
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = clCmResourceEventPublish(node->sessionId, &node->event,&node->rptentry, &node->rdr))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource event publish failed, error [0x%x]", rc);
                }
            }
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
            {
                if ((rc = clCmResourceEventInformAmf(node->sessionId, &node->event,&node->rptentry, &node->rdr))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource event notification to Amf failed, error [0x%x]", rc);
                }
            }

            #if 0
            /*Process the ET Resource event */ 
            if ((rc = clCmResourceEventProcess(node->sessionId, &node->event,&node->rptentry, &node->rdr))!= CL_OK )
            {
                clLog(CL_LOG_ERROR, AREA_HPI, CTX_EVT, "Resource event processing failed, error [0x%x]", rc);
            }
            #endif
            break;

        case  SAHPI_ET_HOTSWAP :
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = resourceHotSwapEventPublish(node->sessionId, &node->event,&node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource Hotswap event publish failed, error [0x%x]", rc);
                }
            }
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
            {
                if ((rc = resourceHotSwapEventInformAmf(node->sessionId, &node->event,&node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource Hotswap event notifyication to amf failed, error [0x%x]", rc);
                }
            }
 
            #if 0
            /*Process the Hotswap event */ 
            if ((rc = resourceHotSwapEventProcess(node->sessionId, &node->event,&node->rptentry))!= CL_OK )
            {
                clLog(CL_LOG_ERROR, AREA_HPI, CTX_EVT, "Hotswap event processing failed, error [0x%x]", rc);
            }
            #endif
            break;   

        case  SAHPI_ET_SENSOR  :
          
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = clCmSensorEventPublish(node->sessionId, &node->event,&node->rptentry, &node->rdr))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Sensor event publish failed, error [0x%x]", rc);
                }
            }
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
            {
                if ((rc = clCmSensorEventInformAmf(node->sessionId, &node->event,&node->rptentry, &node->rdr))!= CL_OK )
                {
                     clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Sensor event notification to Amf failed, error [0x%x]", rc);
                }
            }

            #if 0
            if ((rc = clCmSensorEventProcess(node->sessionId, &node->event, &node->rptentry, &node->rdr)) != CL_OK)
            {
                clLog(CL_LOG_ERROR, AREA_HPI, CTX_EVT, "Sensor event processing failed, error [0x%x]", rc);
            }
            #endif
#ifndef HPI_CHANNEL_EVENT_SIM
            break; 
#endif
        case  SAHPI_ET_DOMAIN:
            clLog(CL_LOG_SEV_DEBUG, AREA_HPI, CTX_EVT,"Replay Domain event for Domain Id [%d] on the Domain Id [%d]", node->event.EventDataUnion.DomainEvent.DomainId, node->domainId);
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = resourceDomainEventPublish(node->sessionId, &node->event, &node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT,"Domain event publish failed, error [0x%x]", rc);
                }
            }
            if (node->pluginOps & CL_CM_PLUGIN_OP_PERMITTED )
            {
                if ((rc = resourceDomainEventProcess(node->sessionId, &node->event, &node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Domain  event process failed, error [0x%x]", rc);
                }
            }
    
#ifndef HPI_CHANNEL_EVENT_SIM
            break;
#endif
        case  SAHPI_ET_WATCHDOG:
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = resourceWatchDogEventPublish(node->sessionId, &node->event, &node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Resource WatchDog event publish failed, error [0x%x]", rc);
                }
            }
#ifndef HPI_CHANNEL_EVENT_SIM
            break;
#endif
        case  SAHPI_ET_OEM:
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = resourceOEMEventPublish(node->sessionId, &node->event, &node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Hotswap event publish failed, error [0x%x]", rc);
                }
            }
#ifndef HPI_CHANNEL_EVENT_SIM
            break;
#endif
        case  SAHPI_ET_USER:
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = resourceUserEventPublish(node->sessionId, &node->event, &node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Hotswap event publish failed, error [0x%x]", rc);
                }
            }
#ifndef HPI_CHANNEL_EVENT_SIM
            break;
#endif
        case  SAHPI_ET_SENSOR_ENABLE_CHANGE:     
            if ( node->pluginOps & CL_CM_PLUGIN_OP_PUBLISHED )
            {
                if ((rc = resourceSensorEnableEventPublish(node->sessionId, &node->event, &node->rptentry))!= CL_OK )
                {
                    clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Hotswap event publish failed, error [0x%x]", rc);
                }
            }

            break;

        default :
            clLog(CL_LOG_SEV_ERROR, AREA_HPI, CTX_EVT, "Invalid event type [%d] received, ignoring", node->event.EventType);
            break;

    }

}
